"""Module rparams of viperleed.calc.classes.rparams.

This is the module defining the core class of this package, i.e.,
Rparams. The class contains parameters read from the PARAMETERS
file, and some parameters defined at runtime. The attributes that
represent not-so-obvious user parameters are instances of classes
defined as part of the special sub-package of rparams. Refactored
on 2023-10-23 by @michele-riva.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-06-13'
__license__ = 'GPLv3+'

from collections import defaultdict
import copy
import logging
import os
from pathlib import Path
import random
import shutil
from timeit import default_timer as timer

try:
    import matplotlib.pyplot as plt
except ImportError:
    _CAN_PLOT = False
else:
    _CAN_PLOT = True

import numpy as np

from viperleed.calc.classes.searchpar import SearchPar
from viperleed.calc.files import beams as iobeams
from viperleed.calc.files.iodeltas import checkDelta
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.base import available_cpu_count
from viperleed.calc.lib.base import parent_name
from viperleed.calc.lib.checksums import KNOWN_TL_VERSIONS
from viperleed.calc.lib.checksums import UnknownTensErLEEDVersionError
from viperleed.calc.sections.calc_section import EXPBEAMS_NAMES
from viperleed.calc.lib.version import Version
from viperleed.calc.files.tenserleed import get_tenserleed_sources

from .defaults import DEFAULTS, NO_VALUE, TENSERLEED_FOLDER_NAME
from .limits import PARAM_LIMITS
from .special.base import NotASpecialParameterError
from .special.base import SpecialParameter


_LOGGER = logging.getLogger(parent_name(__name__))
if _CAN_PLOT:
    plt.style.use('viperleed.calc')


class Rparams:
    """Stores the parameters found in a PARAMETERS file (default values in
    __init__), as well as some parameters defined at runtime."""

    def __init__(self):
        self.readParams = defaultdict(list)  # Raw PARAMETERS from file

        # FROM PARAMETERS FILE
        self.ATTENUATION_EPS = 0.001
        self.AVERAGE_BEAMS = None
        self.BULKDOUBLING_EPS = 0.001
        self.BULKDOUBLING_MAX = 10
        self.BULK_LIKE_BELOW = 0.
        self.BULK_REPEAT = None
        self.DOMAINS = {}         # {name: path_to_tensors_zip_or_dir}
        self.DOMAIN_STEP = 1      # area step in percent for domain search
        self.ELEMENT_MIX = {}     # {element_name: splitlist}
        self.ELEMENT_RENAME = {}  # {element_name: chemical_element}
        self.FILAMENT_WF = DEFAULTS['FILAMENT_WF']['lab6']   # work function of emitting cathode
        self.FORTRAN_COMP = ['', '']      # before files, after files
        self.FORTRAN_COMP_MPI = ['', '']  # before files, after files
        self.GAUSSIAN_WIDTH = DEFAULTS['GAUSSIAN_WIDTH']
        self.GAUSSIAN_WIDTH_SCALING = DEFAULTS['GAUSSIAN_WIDTH_SCALING']
        self.HALTING = 2    # 2: major concerns, 1: minor warnings, 0: always
        self.INTPOL_DEG = 3 # Degree of interpolation spline used in R-factor calculation
        self.IV_SHIFT_RANGE = self.get_default('IV_SHIFT_RANGE')
        self.LAYER_CUTS = self.get_default('LAYER_CUTS')
        self.LAYER_STACK_VERTICAL = True
        self.LMAX = self.get_default('LMAX')
        self.LOG_LEVEL = DEFAULTS['LOG_LEVEL'][NO_VALUE]
        self.LOG_SEARCH = True
        self.N_BULK_LAYERS = 1           # number of bulk layers
        self.N_CORES = 0                 # number of cores
        # OPTIMIZE: settings for fd optimization
        self.OPTIMIZE = self.get_default('OPTIMIZE')
        self.PARABOLA_FIT = {'type': 'none', 'alpha': 1e-2, 'mincurv': 5e-3,
                             'localize': 0}
        self.PHASESHIFT_EPS = DEFAULTS['PHASESHIFT_EPS']['d'] # changed in updateDerivedParams
        self.PHASESHIFTS_CALC_OLD = True # use old EEASiSSS version # TODO: once established, set to False or remove
        self.PHASESHIFTS_OUT_OLD = True  # output old PHASESHIFTS file # TODO: once established, set to False or remove
        self.PHI = DEFAULTS['PHI']           # from BEAM_INCIDENCE
        self.PLOT_IV = {'plot': True, 'axes': 'all', 'colors': [],
                        'font_size': 10, 'legend': 'all', 'line_width': 1,
                        'overbar': False, 'perpage': 2,
                        }
        self.RUN = self.get_default('RUN')        # what segments should be run
        self.R_FACTOR_LEGACY = True # use old runtime-compiled R-factor calculation
        self.R_FACTOR_TYPE = 1  # 1: Pendry, 2: R2, 3: Zanazzi-Jona
        self.R_FACTOR_SMOOTH = 0
        self.S_OVL = 0.3 # Muffin tin overlap parameter after Rundgren 2021, default is 0.3 - set or optimize in FD
        self.SCREEN_APERTURE = 110.
        self.SEARCH_BEAMS = 0   # 0: average, 1: integer, 2: fractional
        # SEARCH_CULL: fraction of population, or absolute nr. if >1
        # SEARCH_CULL.type_: clone, genetic, random
        self.SEARCH_CULL = self.get_default('SEARCH_CULL')
        self.SEARCH_MAX_GEN = 100000  # maximum number of generations in search
        self.SEARCH_MAX_DGEN = self.get_default('SEARCH_MAX_DGEN')
        # maximum number of generations without change before search
        #   is stopped. All: all configs, best: only 1, dec: best 10%
        #    0: don't use parameter
        self.SEARCH_MAX_DGEN_SCALING = {'all': None, 'best': None, 'dec': None}
        # will be defined in updateDerivedParams
        self.SEARCH_LOOP = False
        self.SEARCH_POPULATION = 0  # trial structures in search
        self.SEARCH_START = 'crandom'
        self.SITE_DEF = {}   # {element_name: {sitename, [atom.num]}}
        self.STOP = False
        self.SUPERLATTICE = np.identity(2, dtype=float)
        self.SUPPRESS_EXECUTION = False
        self.SYMMETRIZE_INPUT = True
        self.SYMMETRY_CELL_TRANSFORM = np.identity(2, dtype=float)
        self.SYMMETRY_EPS = self.get_default('SYMMETRY_EPS')
        self.SYMMETRY_FIND_ORI = True
        self.SYMMETRY_FIX = self.get_default('SYMMETRY_FIX')
        self.SYMMETRY_BULK = {}   # keys: group, rotation, mirror
        self.TENSOR_INDEX = None  # default: pick highest in Tensors folder
        self.TENSOR_OUTPUT = []  # per layer: write Tensor output? (0/1)
        # THEO_ENERGIES: the default values without experimental
        # beams is set in section INIT via self.initTheoEnergies
        self.THEO_ENERGIES = self.get_default('THEO_ENERGIES')
        self.THETA = DEFAULTS['THETA']        # from BEAM_INCIDENCE
        self.TL_IGNORE_CHECKSUM = False
        self.TL_VERSION = self.get_default('TL_VERSION')
        self.T_EXPERIMENT = None
        self.T_DEBYE = None
        self.V0_IMAG = 4.5
        self.V0_REAL = 'default'   # 'default' will read from PHASESHIFTS
        self.V0_Z_ONSET = 1.0
        self.VIBR_AMP_SCALE = []   # read as list of strings, interpret later
        self.ZIP_COMPRESSION_LEVEL = DEFAULTS['ZIP_COMPRESSION_LEVEL']

        # RUN VARIABLES
        self.starttime = timer()
        self.source_dir = None  # where to find 'tensorleed'
        # .workdir is the MAIN WORK DIRECTORY; where to find input
        self.workdir = Path.cwd().resolve()
        self.compile_logs_dir = None
        self.searchConvInit = {
            'gaussian': None, 'dgen': {'all': None, 'best': None, 'dec': None}}
        self.searchEvalTime = DEFAULTS['SEARCH_EVAL_TIME']  # time interval for reading SD.TL
        self.output_interval = None # changed in updateDerivedParams
        self.searchMaxGenInit = self.SEARCH_MAX_GEN
        self.searchStartInit = None

        # script progress tracking
        self.halt = 0
        self.systemName = ''
        self.timestamp = ''
        self.manifest = ['SUPP', 'OUT']
        self.fileLoaded = {
            'PARAMETERS': True, 'POSCAR': False,
            'IVBEAMS': False, 'VIBROCC': False, 'PHASESHIFTS': False,
            'DISPLACEMENTS': False, 'BEAMLIST': False, 'EXPBEAMS': False
            }
        self.runHistory = []   # sections that have been executed before
        self.lastOldruns = []
        # copy of runHistory when last oldruns folder was created
        self.superlattice_defined = False
        self.ivbeams_sorted = False
        self.last_R = None
        self.stored_R = {'refcalc': None, 'superpos': None}
        self.checklist = []  # output strings of things to check at program end

        # domains
        self.domainParams = []
        self.pseudoSlab = None

        # data from files
        self.beamlist = []  # lines as strings from BEAMLIST
        self.ivbeams = []   # uses Beam class; list of beams only
        self.expbeams = []  # uses Beam class; contains intensities
        self.theobeams = {  # uses Beam class; contains intensities
            'refcalc': None,
            'superpos': None
            }
        self.expbeams_file_name = ''     # EXPBEAMS or EXPBEAMS.csv?
        self.phaseshifts = []
        self.phaseshifts_firstline = ''  # contains parameters for MUFTIN
        self.refcalc_fdout = ''
        self.superpos_specout = ''
        self.best_v0r = None     # best value for v0r from previous R-factor
        self.disp_blocks = []    # tuples (lines, name) in DISPLACEMENTS file
        self.disp_block_read = False  # current displacements block read?
        self.disp_loops = []          # list of tuples (loopStart, loopEnd)
        self.controlChemBackup = None

        # search parameters
        self.searchpars = []
        self.searchResultConfig = None
        self.search_atlist = []    # atoms that are relevant for the search
        self.search_maxfiles = 0   # maximum number of delta files for one atom
        self.search_maxconc = 1    # maximum number of concentration steps
        self.indyPars = 0          # number of independent parameters
        self.mncstep = 0    # max. steps (geo. times therm.) for one atom
        self.search_index = 0      # which DISPLACEMENTS block is being done

        # plotting data
        self.rfacscatter_all = []   # tuples (gen, r, size, color)
        self.rfacscatter = []       # same, but thinned out along gens
        self.parScatter = [[]]
        # tuples (gen, mean scatter, max scatter) per search
        self.searchplots = [('', [], [], [], [])]
        # (name, gens, min, max, mean) for each search
        self.lastParScatterFigs = {}
        # complete figures for each search, with search names as keys

    @property
    def is_debug_mode(self):
        return self.LOG_LEVEL < logging.INFO

    @property
    def no_value(self):
        return NO_VALUE

    def try_loading_expbeams_file(self):
        """Read an EXPBEAMS file if not already available."""
        if self.fileLoaded['EXPBEAMS']:
            return

        exp_files_provided = (f for f in EXPBEAMS_NAMES if Path(f).is_file())
        try:
            self.expbeams_file_name = next(exp_files_provided)
        except StopIteration:  # Nothing to load
            self.expbeams_file_name = ''
            return

        # Warn if multiple experimental input files were provided
        if next(exp_files_provided, False):
            self.setHaltingLevel(1)
            _LOGGER.warning(
                'Multiple files with experimental I(V) curves were provided. '
                'Check if the root directory contains the correct files. '
                f'Using file {self.expbeams_file_name!r}.'
                )
        err_msg = f'Error while reading file {self.expbeams_file_name}'
        enrange = self.THEO_ENERGIES.as_floats()[:2]
        try:
            self.expbeams = iobeams.readOUTBEAMS(self.expbeams_file_name,
                                                 enrange=enrange)
        except OSError:
            _LOGGER.error(f'{err_msg}.', exc_info=True)
            return

        if self.expbeams:
            self.fileLoaded['EXPBEAMS'] = True
        else:
            _LOGGER.error(f'{err_msg}: No data was read.')

    @classmethod
    def get_default(cls, param):
        """Return the default value of param."""
        try:
            value = DEFAULTS[param]
        except KeyError as err:
            raise ValueError(f'No default found for parameter {param}') from err
        if isinstance(value, tuple):
            value = list(value)
        elif isinstance(value, dict):
            value = copy.deepcopy(value)

        # See if param is a special one
        return cls._to_simple_or_special_param(param, value)

    @staticmethod
    def _to_simple_or_special_param(param, value):
        """Return the value of a standard or special parameter."""
        try:
            special = SpecialParameter.get_subclass(param)
        except NotASpecialParameterError:
            return value
        return special.from_value(value)

    def reset_default(self, param):
        """Reset param to its default value."""
        default = self.get_default(param)
        setattr(self, param, default)

    @staticmethod
    def get_limits(param):
        """Return the smallest and largest acceptable values of param."""
        try:
            return PARAM_LIMITS[param]
        except KeyError as err:
            raise ValueError(f'No limits found for parameter {param}') from err

    def update(self, presets):
        """Load presets into this Rparams object."""
        for param_name, param_value in presets.items():
            param_name = param_name.upper()
            getattr(self, param_name)  # Raise AttributeError if wrong
            value = self._to_simple_or_special_param(param_name, param_value)
            setattr(self, param_name, value)

    def total_energy_range(self):
        """Return the total overlapping energy range of experiment and
        theory. Note that this may change if experimental beams are dropped."""
        if not self.expbeams:
            return 0.
        totalrange = 0.
        for b in self.expbeams:
            totalrange += (min(max(b.intens), self.THEO_ENERGIES.max)
                           - max(min(b.intens), self.THEO_ENERGIES.min))
        return totalrange

    def storeRfacScatter(self, x, y, s, c):
        """
        Adds a list of points for r-factor scatter plots to
        self.rfacscatter

        Parameters
        ----------
        x, y, s, c : lists of floats
            coordinates, size and color of points.

        Returns
        -------
        None.
        """
        spacing = x[-1]/50
        pg = x[-1]
        self.rfacscatter_all.extend(list(zip(x, y, s, c)))
        self.rfacscatter = []
        for p in self.rfacscatter_all[::-1]:
            if p[0] <= pg - spacing or p[0] == pg:
                self.rfacscatter.append(p)
                pg = p[0]

    def get_tenserleed_directory(self, wanted_version=None):                           # TODO: replace the default for TL_VERSION with Version('unknown')
        """Return the Path to a TensErLEED directory.

        The directory is looked up in Rparams.source_dir.

        Parameters
        ----------
        version : Version, str, optional
            Which specific version of TensErLEED should be looked
            up. If not given or None, `Rparams.TL_VERSION` is used.
            If `version == 0`, the highest version is returned.
            Default is None.

        Returns
        -------
        tenserleed_path : Path
            Path to the desired 'TensErLEED' directory or zip archive.

        Raises
        ------
        RuntimeError
            If this method is called before `Rparams.source_dir` is set
        FileNotFoundError
            If `Rparams.source_dir` has no 'TensErLEED' subdirectories
        FileNotFoundError
            If `version` is given, but the corresponding directory
            was not found
        """
        if not self.source_dir:
            raise RuntimeError(
                f'{type(self).__name__}.get_tenserleed_directory: '
                'source_dir is not set'
                )
        source_tree = self.source_dir.resolve()
        wanted_version = (Version(wanted_version) if wanted_version else 
                          self.TL_VERSION)
        sources = get_tenserleed_sources(source_tree)
        # sort sources by .version attribute
        sources = sorted(sources, key=lambda x: x.version)

        if wanted_version is None:
            # return highest version
            return max(sources, key=lambda x: x.version)

        # otherwise look for the wanted version
        for source in sources:
            if source.version == wanted_version:
                return source
        raise FileNotFoundError(
            f'Could not find TensErLEED version={wanted_version} in '
            f'{source_tree}.'
            )


    def updateDerivedParams(self):
        """
        Checks which derivative parameters (which cannot be calculated at
        initialization) can be calculated now

        Returns
        -------
        None.
        """
        # TENSOR_INDEX:
        if self.TENSOR_INDEX is None:
            self.TENSOR_INDEX = leedbase.getMaxTensorIndex()
        # TL_VERSION:
        if self.source_dir is None:
            raise RuntimeError('Cannot determine highest TensErLEED version '
                               'without specifying a source directory')
        if self.TL_VERSION is None:

            # Fetch most recent TensErLEED version
            tl_source = self.get_tenserleed_directory()
            self.TL_VERSION = tl_source.version
            _LOGGER.debug(f'Detected TensErLEED version {str(self.TL_VERSION)}')


        # SEARCH_CONVERGENCE:
        if self.searchConvInit['gaussian'] is None:
            self.searchConvInit['gaussian'] = self.GAUSSIAN_WIDTH
        for k in ['all', 'best', 'dec']:
            if self.SEARCH_MAX_DGEN_SCALING[k] is None:
                self.SEARCH_MAX_DGEN_SCALING[k] = 1 / self.GAUSSIAN_WIDTH_SCALING
            if self.searchConvInit['dgen'][k] is None:
                self.searchConvInit['dgen'][k] = self.SEARCH_MAX_DGEN[k]
        if self.output_interval is None:
            # set output interval to SEARCH_CONVERGENCE dgen, but cap at 1000
            use_dgen = min(dgen for dgen in self.searchConvInit['dgen'].values() if dgen > 0) or 1000
            self.output_interval = int(min(use_dgen or 1000, 1000))  # default to 1000 if all dgen are 0 (default)
        if self.searchStartInit is None:
            self.searchStartInit = self.SEARCH_START
        # Phaseshifts-based:
        if self.fileLoaded['PHASESHIFTS']:
            # get highest required energy index
            hi = len(self.phaseshifts)-1
            if self.THEO_ENERGIES.max is not NO_VALUE:
                for i in range(0, len(self.phaseshifts)):
                    if (self.phaseshifts[i][0]*leedbase.HARTREE_TO_EV
                            > self.THEO_ENERGIES.max):
                        hi = i
                        break
            # LMAX
            if self.PHASESHIFT_EPS == 0:
                self.PHASESHIFT_EPS = DEFAULTS['PHASESHIFT_EPS']['f']
            min_set = self.LMAX.has_min
            if not min_set:
                self.LMAX.min = 6
            if not self.LMAX.has_max:  # determine value from PHASESHIFT_EPS
                lmax = 1
                for el in self.phaseshifts[hi][1]:  # only check highest energy
                    for i, val in enumerate(el):
                        if abs(val) > self.PHASESHIFT_EPS and (i+1) > lmax:
                            lmax = i+1
                if lmax < 8 and not min_set:
                    _LOGGER.debug('Found small LMAX value based on '
                                  f'PHASESHIFT_EPS parameter (LMAX={lmax})')
                if lmax > 18:
                    lmax = 18
                    _LOGGER.info(
                        'The LMAX found based on the PHASESHIFT_EPS '
                        'parameter is greater than 18, which is currently '
                        'not supported. LMAX was set to 18.'
                        )
                self.LMAX.max = lmax
            else:       # sanity check: are large values ignored?
                warn = False
                highval = 0
                for el in self.phaseshifts[hi][1]:   # highest energy
                    for i, val in enumerate(el):
                        if abs(val) > 0.1 and (i+1) > self.LMAX.max:
                            warn = True
                            highval = max(highval, abs(val))
                if warn:
                    _LOGGER.warning(
                        'The LMAX value defined in the PARAMETERS '
                        'file leads to large phaseshift values being ignored '
                        f'(highest value: {highval}). Consider using a higher '
                        'LMAX, or defining LMAX indirectly via PHASESHIFT_EPS.'
                        )
                    self.setHaltingLevel(1)
            # V0_REAL
            if self.V0_REAL == 'default':
                llist = self.phaseshifts_firstline.split()
                c = []
                try:
                    for i in range(0, 4):
                        c.append(float(llist[i + 1]))
                except ValueError:
                    _LOGGER.error('Could not read Muftin parameters from '
                                  'PHASESHIFTS file.')
                    raise
                self.V0_REAL = c

    def updateCores(self):
        """
        If self.N_CORES is undefined, tries to find it

        Raises
        ------
        RuntimeError
            If unable to detect number of cores.

        Returns
        -------
        None
        """
        if self.N_CORES:
            return
        try:
            self.N_CORES = available_cpu_count()
        except Exception:
            _LOGGER.error('Failed to detect number of cores.')
            raise

        _LOGGER.info('Automatically detected number '
                     f'of available CPUs: {self.N_CORES}')
        if not self.N_CORES:
            _LOGGER.error('Failed to detect number of cores.')
            raise RuntimeError('N_CORES undefined, automatic detection failed')

    def resetSearchConv(self):
        """
        Resets the search convergence and tracking parameters to their
        initial values.

        Returns
        -------
        None.
        """
        self.controlChemBackup = None
        self.disp_block_read = False
        self.rfacscatter_all = []
        self.searchplots.append(('', [], [], [], []))
        self.parScatter.append([])
        self.SEARCH_MAX_GEN = self.searchMaxGenInit
        if self.searchConvInit['gaussian'] is not None:
            self.GAUSSIAN_WIDTH = self.searchConvInit['gaussian']
        if self.searchStartInit is not None:
            self.SEARCH_START = self.searchStartInit
        if self.SEARCH_START == 'control':
            self.SEARCH_START = 'crandom'
        for k in ['best', 'all', 'dec']:
            if self.searchConvInit['dgen'][k] is not None:
                self.SEARCH_MAX_DGEN[k] = self.searchConvInit['dgen'][k]

    def setHaltingLevel(self, set_to):
        """
        Sets the halting level self.halt to value set_to, if that value is
        higher than the current halting level. Outputs debug message.

        Parameters
        ----------
        set_to : int
            New value for self.halt

        Returns
        -------
        None.
        """
        if set_to > self.halt:
            _LOGGER.debug(f'Raising halting level to {set_to}')
            self.halt = set_to

    def initTheoEnergies(self):
        """
        Initializes values for the THEO_ENERGIES parameter either to default,
        or if experimental beams were loaded to values corresponding to the
        experimental energy range.

        Returns
        -------
        None.
        """
        energies = self.THEO_ENERGIES
        if energies.defined:
            return
        info = self.fileLoaded['EXPBEAMS'] and not energies.has_bounds
        if self.fileLoaded['EXPBEAMS']:
            minen = min(en for beam in self.expbeams for en in beam.intens)
            maxen = max(en for beam in self.expbeams for en in beam.intens)
            values = [minen - 3, maxen + 3, 3]
        else:
            values = self.get_default('THEO_ENERGIES - no experiments')

        energies.set_undefined_values(values)
        if info:
            _LOGGER.debug('Initialized energy range from experimental '
                          f'beams file: {energies.start:.2f} to '
                          f'{energies.stop:.2f} eV')

    # TODO: eventually, these default values should be moved to some constant or other file
    def getFortranComp(self, comp='auto'):                                      # TODO: combine with FortranCompMPI from below; lots of repeated code
        """
        Checks whether ifort or gfortran are present, and sets FORTRAN_COMP
        accordingly.

        Parameters
        ----------
        comp : str, optional
            'auto' (default): will check ifort first, then gfortran.
            'ifort' or 'gfortran': only that compiler will be checked.

        Raises
        ------
        ValueError
            If the given compiler is not supported
        FileNotFoundError
            If the desired compiler was not found
        """
        # supported compilers, in order of priority
        supported = ['ifort', 'gfortran']
        found = ''
        if comp == 'auto':
            check = supported
        elif comp in supported:
            check = [comp]
        else:
            _LOGGER.error('Rparams.getFortranComp: requested compiler is not '
                          'supported.')
            raise ValueError('Fortran compiler not supported')
        for c in check:
            if shutil.which(c, os.X_OK) is not None:
                found = c
                break
        if found == '':
            if comp == 'auto':
                _LOGGER.error('Rparams.getFortranComp: No fortran compiler '
                              'found.')
            else:
                _LOGGER.error('Rparams.getFortranComp: Requested fortran '
                              'compiler not found.')
            raise FileNotFoundError('Fortran compiler not found')
        if found == 'ifort':
            self.FORTRAN_COMP = [
                'ifort -O2 -I/opt/intel/mkl/include',
                '-L/opt/intel/mkl/lib/intel64 -lmkl_intel_lp64 '
                '-lmkl_intel_thread -lmkl_core -liomp5 -lpthread -lm -ldl '
                '-traceback']  # backtrace should not affect performance
            _LOGGER.debug('Using fortran compiler: ifort')
        elif found == 'gfortran':
            self.FORTRAN_COMP = ['gfortran -O2', '-llapack -lpthread -lblas '
                                 '-fbacktrace']  # should not affect performance
            _LOGGER.debug('Using fortran compiler: gfortran')

    def getFortranMpiComp(self, comp='auto'):
        """
        Checks whether mpiifort or mpifort are present, and sets FORTRAN_COMP
        mpi accordingly.

        Parameters
        ----------
        comp : str, optional
            'auto' (default): will check mpiifort first, then mpifort.
            'mpiifort' or 'mpifort': only that compiler will be checked.

        Raises
        ------
        ValueError
            If the given compiler is not supported
        FileNotFoundError
            If the desired compiler was not found
        """
        # supported compilers, in order of priority
        supported = ['mpiifort', 'mpifort']
        found = ''
        if comp == 'auto':
            check = supported
        elif comp in supported:
            check = [comp]
        else:
            _LOGGER.error('Rparams.getFortranMpiComp: requested compiler is '
                          'not supported.')
            raise ValueError('Fortran MPI compiler not supported')
        for c in check:
            if shutil.which(c, os.X_OK) is not None:
                found = c
                break
        if found == '':
            if comp == 'auto':
                _LOGGER.error('Rparams.getFortranMpiComp: No fortran compiler '
                              'found.')
            else:
                _LOGGER.error('Rparams.getFortranMpiComp: Requested fortran '
                              'compiler not found.')
            raise FileNotFoundError('Fortran MPI compiler not found')
        if found == 'mpiifort':
            self.FORTRAN_COMP_MPI = ['mpiifort -Ofast', '']
            _LOGGER.debug('Using fortran compiler: mpiifort')
        elif found == 'mpifort':
            self.FORTRAN_COMP_MPI = [
                'mpifort -Ofast -no-pie -fallow-argument-mismatch',
                ''
                ]
            _LOGGER.debug('Using fortran compiler: mpifort')

    def renormalizeDomainParams(self, config):
        """Takes a list of parameter indices as produced by e.g.
        getRandomConfig, and checks it for domain parameters. If the domain
        parameters can be multiplied by an integer value and still be in
        range, returns the list with multiplied values; else, returns the
        unchanged list."""
        domain_indices = [i for (i, sp) in enumerate(self.searchpars)
                          if sp.mode == 'dom']
        if not domain_indices or all(config[i] == 1 for i in domain_indices):
            return config
        mult = 1
        domain_steps = self.searchpars[domain_indices[0]].steps
        while (sum(config[i]-1 for i in domain_indices) * (mult + 1)
               <= domain_steps-1):
            mult += 1
        for i in domain_indices:
            config[i] = ((config[i] - 1) * mult) + 1
        return config

    def getCenteredConfig(self):
        """Returns a list of 'centered' parameter indices, i.e. all as close
        to 'no displacement' as possible."""
        return [sp.center for sp in self.searchpars]

    def getPredictConfig(self, best_config=None, curv_cutoff=1e-4):
        """
        Outputs a parameter configuration as list of integers that is as close
        to the result of the parabola fit as possible. For parameters with no
        valid parabola fit result, or curvatures below the cutoff, the value
        from the best known configuration will be used instead. If the best
        known configuration is not passed, these parameters will instead
        assume the value in the center of their displacement range.

        Parameters
        ----------
        best_config : list of int, optional
            The current best configuration, to be used for parameters that
            do not have usable parabola fit results. If not passed, centered
            values will be used instead.
        curv_cutoff : float, optional
            The minimum curvature that a parameter parabola needs to have in
            order to use the parabola minimum. The default is 1e-4.

        Returns
        -------
        list of int
            Parameter values for a new configuration.
        """
        out = []
        for (i, sp) in enumerate(self.searchpars):
            if (sp.parabolaFit['min'] is not None and
                    not np.isnan(sp.parabolaFit['err_co']) and
                    not np.isnan(sp.parabolaFit['err_unco'])):
                # sp.parabolaFit['curv'] is not None and
                # sp.parabolaFit['curv'] > curv_cutoff):
                out.append(int(round(sp.parabolaFit['min'])))
            else:
                if best_config is not None:
                    out.append(best_config[i])
                else:
                    out.append(int((sp.steps + 1)/2))
        return self.renormalizeDomainParams(out)

    def getRandomConfig(self):
        """
        Outputs a new random parameter configuration. Linked parameters are
        guaranteed to have the same value.

        Returns
        -------
        list of int
            Parameter values for a new configuration.
        """
        out = [-1] * len(self.searchpars)
        for i, sp in enumerate(self.searchpars):
            out[i] = random.randint(1, sp.steps)
        for i, sp in enumerate(self.searchpars):
            if sp.linkedTo is not None:
                out[i] = out[self.searchpars.index(sp.linkedTo)]
            elif sp.restrictTo is not None:
                if type(sp.restrictTo) == int:
                    out[i] = sp.restrictTo
                else:
                    out[i] = out[self.searchpars.index(sp.restrictTo)]
        if -1 in out:
            _LOGGER.error(f'Rparams.getRandomConfig failed: {out}')
            return []
        return self.renormalizeDomainParams(out)

    def getOffspringConfig(self, parents):
        """
        Returns a list of parameter indices generated as a random mix of
        the parameters from two of the configurations passed as 'parents',
        picked at random if there are more than two. If possible, the parents
        will be picked such that they are not identical, but the offspring may
        nevertheless be identical with one of the parents.

        Parameters
        ----------
        parents : list of list of int
            List of surviving configurations.

        Returns
        -------
        list of int
            Parameter values for a new configuration.
        """
        if not parents:
            _LOGGER.warning('Rparams.getOffspringConfig: Cannot create '
                            'offspring configuration without parents. '
                            'Returning random configuration')
            return self.getRandomConfig()

        if len(parents) == 1:
            _LOGGER.warning('Rparams.getOffspringConfig: Only one parent '
                            'passed. Returning clone.')
            return parents[0]

        parents = parents[:]
        i = random.randint(0, len(parents)-1)
        p2 = [parents.pop(i)]
        while len(p2) == 1:
            if len(parents) == 1:
                p2.append(parents[0])
                break
            i = random.randint(0, len(parents)-1)
            p = parents.pop(i)
            if p2[0] != p:
                p2.append(p)
        out = [-1] * len(self.searchpars)
        for i, sp in enumerate(self.searchpars):
            if sp.restrictTo is None and sp.linkedTo is None:
                out[i] = p2[random.randint(0, 1)][i]
        for i, sp in enumerate(self.searchpars):
            if sp.restrictTo is not None:
                if type(sp.restrictTo) == int:
                    out[i] = sp.restrictTo
                else:
                    out[i] = out[self.searchpars.index(sp.restrictTo)+1]
            elif sp.linkedTo is not None:
                out[i] = out[self.searchpars.index(sp.linkedTo)+1]
        if -1 in out:
            _LOGGER.error(f'Rparams.getOffspringConfig failed: {out}')
            return []
        return self.renormalizeDomainParams(out)

    def closePdfReportFigs(self):
        """
        Closes the pdf figures from the Search-progress pdf files, which are
        kept in memory for the Search-report file during a run.

        Returns
        -------
        None.
        """
        if not _CAN_PLOT:
            return

        for figures in self.lastParScatterFigs.values():
            for figure in figures:
                try:
                    plt.close(figure)
                except Exception:
                    pass

    def generateSearchPars(self, sl, subdomain=False):
        """
        Initializes a list of Searchpar objects, and assigns delta files to
        atoms if required. Also sets the self.indyPars parameter.

        Parameters
        ----------
        sl : Slab
            The Slab object containing atom information.
        subdomain : bool, optional
            Set True if executing for multiple domains, and this is not the
            highest-level Rprarams object.

        Raises
        ------
        ValueError
            If inconsistent values are found.
        RuntimeError
            If required files are missing.

        Returns
        -------
        None.
        """
        if self.domainParams:
            self.generateSearchPars_domains()
            return

        self.searchpars = []
        self.search_maxfiles = 0   # maximum number of delta files for one atom
        self.search_maxconc = 1
        self.indyPars = 0          # number of independent parameters
        self.mncstep = 0     # max. steps (geo. times therm.) for one atom
        # track which atoms are symmetry-linked to the ones already done to
        #   not double-count indyPars
        eqlist = []
        # get list of atoms that appear in the search
        if (2 in self.runHistory or 42 in self.runHistory
                or sl.deltas_initialized):
            # if delta has been run, information what deltas exist is stored
            atlist = [at for at in sl if not at.is_bulk and at.known_deltas]
        else:
            _LOGGER.debug('Delta-amplitudes were not calculated in current '
                          'run; looking for delta files by name.')
            # see what delta files are present
            filelist = [filename for filename in os.listdir('.') if
                        filename.startswith('DEL_')]
            delN = []
            for filename in filelist:
                try:
                    delN.append(int(filename.split('_')[1]))
                except ValueError:
                    filelist.remove(filename)
            atlist = [at for at in sl if not at.is_bulk and at.num in delN]
            for at in atlist:
                deltaCandidates = [fn for fn in filelist
                                   if int(fn.split('_')[1]) == at.num]
                checkEls = list(at.disp_occ.keys())
                # check for vacancy:
                occlists = []
                for k in at.disp_occ:
                    occlists.append(at.disp_occ[k])
                for i in range(0, len(occlists[0])):
                    totalocc = 0.
                    for ol in occlists:
                        if len(ol) <= i:
                            _LOGGER.error(
                                f'Inconsistent occupancy lists for {at}'
                                )
                            raise ValueError('Inconsistent input')
                        else:
                            totalocc += ol[i]
                    if totalocc < 1 - 1e-4:
                        checkEls.append('vac')
                        break
                # now check if all deltas are there:
                for el in checkEls:
                    found = False
                    for df in [f for f in deltaCandidates
                               if f.split('_')[2] == el]:
                        if checkDelta(df, at, el, self):
                            found = True
                            at.known_deltas.append(df)
                            break
                    if not found:
                        _LOGGER.error('No appropriate Delta file found '
                                      f'for {at}, element {el}')
                        raise RuntimeError('Missing Delta file')
            # sanity check: are displacements defined but deltas missing?
            for at in sl:
                # check whether at has non-trivial displacements:
                found = False
                for d in [at.disp_occ, at.disp_geo, at.disp_vib]:
                    for el in d:
                        if len(d[el]) > 1:
                            found = True
                            break
                if not found:
                    for el in at.disp_vib:
                        if at.disp_vib[el][0] != 0.0:
                            found = True
                            break
                if not found:
                    for el in at.disp_geo:
                        if np.linalg.norm(at.disp_geo[el][0]) >= 1e-4:
                            found = True
                            break
                if not found:
                    occlists = []
                    for k in at.disp_occ:
                        occlists.append(at.disp_occ[k])
                    for i in range(0, len(occlists[0])):
                        totalocc = 0.
                        for ol in occlists:
                            if len(ol) <= i:
                                break  # error - will pop up again later...
                            else:
                                totalocc += ol[i]
                        if totalocc < 1 - 1e-4:
                            found = True
                            break
                if found and at not in atlist:
                    _LOGGER.error(f'{at} has displacements defined, but no '
                                  'delta file was found! Run Delta-Amplitudes.')
                    raise RuntimeError('Delta file not found')
                elif not found and at in atlist:
                    # delta file is there, but no displacements
                    atlist.remove(at)
            sl.deltas_initialized = True
        # sort atlist by displists
        al2 = []
        for dl in sl.displists:
            al2.extend([a for a in atlist if a in dl])
        al2.extend([a for a in atlist if a not in al2])
        atlist = al2
        md = {'geo': 1, 'vib': 2, 'occ': 3}
        splToRestrict = []
        indep = []
        for at in atlist:
            if len(at.known_deltas) > self.search_maxfiles:
                self.search_maxfiles = len(at.known_deltas)
            for fn in at.known_deltas:
                el = fn.split('_')[2]
                if el == 'vac':
                    self.searchpars.append(SearchPar(at, 'geo', 'vac', fn))
                    continue
                mult = 1
                pars = 0
                for (mode, d) in [('vib', at.disp_vib),
                                  ('geo', at.disp_geo)]:
                    if el in d:
                        k = el
                    else:
                        k = 'all'
                    if len(d[k]) > 1 or (len(d[k]) == 1 and
                                         ((mode == 'geo'
                                          # and np.linalg.norm(d[k][0]) >= 1e-4
                                           )
                                          or (mode == 'vib' and
                                                      d[k][0] != 0.))):
                        pars += 1
                        sp = SearchPar(at, mode, el, fn)
                        self.searchpars.append(sp)
                        if el in at.constraints[md[mode]]:
                            k2 = el
                        else:
                            k2 = 'all'
                        if k2 in at.constraints[md[mode]]:
                            c = at.constraints[md[mode]][k2]
                            if type(c) == int:
                                sp.restrictTo = c + 1
                            else:
                                splToRestrict.append((sp, c))
                        elif len(d[k]) > 1 and at not in eqlist:
                            self.indyPars += 1
                            indep.append(sp)
                        if len(d[k]) > 1 and at in eqlist:
                            spl = [s for s in self.searchpars
                                   if at in s.atom.displist
                                   and s.mode == sp.mode
                                   and (s.el == el
                                        or el in ['', 'all'])]
                            if spl:
                                sp.linkedTo = spl[0]
                    mult *= len(d[k])
                if pars == 0:
                    self.searchpars.append(SearchPar(at, 'geo', el, fn))
                if mult > self.mncstep:
                    self.mncstep = mult
            sp = SearchPar(at, 'occ', '', fn)
            self.searchpars.append(sp)
            occsteps = len(next(iter(at.disp_occ.values())))
            if occsteps > 1:
                if occsteps > self.search_maxconc:
                    self.search_maxconc = occsteps
                if at.constraints[3]:
                    c = list(at.constraints[3].values())[0]
                    if type(c) == int:
                        sp.restrictTo = c + 1
                    else:
                        splToRestrict.append((sp, c))
                else:
                    if at not in eqlist:  # occupation will actually vary
                        self.indyPars += 1
                        indep.append(sp)
                if at in eqlist:
                    spl = [s for s in self.searchpars if at
                           in s.atom.displist and s.mode == 'occ']
                    if spl:
                        sp.linkedTo = spl[0]
            eqlist.extend(at.displist)  # do not consider for future indyPars
        splTargets = set()
        for (sp, (at, el)) in splToRestrict:
            # ind = self.searchpars.index(sp)
            spl = [s for s in self.searchpars if s != sp and s.atom == at and
                   s.mode == sp.mode and (s.el == el or
                                          el in ['', 'all'] or
                                          sp.mode == 'occ')]
            if spl:
                sp.restrictTo = spl[0]
                splTargets.add(spl[0])
                if spl[0] not in indep and spl[0].steps > 1:
                    self.indyPars += 1
                    indep.append(spl[0])
        for (sp, (at, el)) in [tup for tup in splToRestrict
                               if tup[0].restrictTo is None
                               and tup[0] not in splTargets]:
            _LOGGER.warning(
                f'Restricting search parameter for atom {sp.atom.num}, '
                f'element {sp.el}, mode {sp.mode} failed: Could not identify '
                f'target search parameter (atom {at.num}, element {el}).'
                )
            if sp not in indep and sp.steps > 1:
                self.indyPars += 1
                indep.append(sp)
        for (i, sp) in enumerate(self.searchpars):
            # restrict to lowest number index, resolve conflicts
            if sp.restrictTo not in self.searchpars:
                continue
            sp2 = None
            while sp2 != sp.restrictTo:
                sp2 = sp.restrictTo
                ind = self.searchpars.index(sp2)
                if (sp2.linkedTo in self.searchpars
                        and self.searchpars.index(sp2.linkedTo) < ind):
                    sp.restrictTo = sp2.linkedTo
                elif (sp2.restrictTo in self.searchpars
                      and self.searchpars.index(sp2.restrictTo) < ind):
                    sp.restrictTo = sp2.restrictTo
            if self.searchpars.index(sp.restrictTo) >= i:
                if sp.restrictTo.restrictTo is None:
                    sp.restrictTo.restrictTo = sp   # invert direction
                if type(sp.restrictTo.restrictTo) == int:
                    sp.restrictTo = sp.restrictTo.restrictTo
                else:
                    sp.restrictTo = None  # remove references to higher indices
        self.search_atlist = atlist
        if not subdomain:
            self.searchpars.append(SearchPar(None, 'dom', '', ''))
            self.searchpars[-1].steps = 2

    def generateSearchPars_domains(self):
        """Call generateSearchPars for every domain and collate results."""
        self.searchpars = []
        self.indyPars = len(self.domainParams) - 1
        home = os.getcwd()
        for dp in self.domainParams:
            try:
                os.chdir(dp.workdir)
                dp.rp.generateSearchPars(dp.sl, subdomain=True)
            except Exception:
                _LOGGER.error('Error while creating delta '
                              f'input for domain {dp.name}')
                raise
            finally:
                os.chdir(home)
            for sp in dp.rp.searchpars:
                if not isinstance(sp.restrictTo, int):
                    continue
                sp.restrictTo += len(self.searchpars)
            self.searchpars.extend(dp.rp.searchpars)
            self.indyPars += dp.rp.indyPars
        for dp in self.domainParams:
            self.searchpars.append(SearchPar(None, 'dom', '', ''))
            self.searchpars[-1].steps = int(100 / self.DOMAIN_STEP) + 1
        self.search_maxfiles = max(dp.rp.search_maxfiles
                                   for dp in self.domainParams)
        self.search_maxconc = max(dp.rp.search_maxconc
                                  for dp in self.domainParams)
        self.mncstep = max(dp.rp.mncstep for dp in self.domainParams)



"""LEED-specific functions used throughout the viperleed calc package."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-06-13'
__license__ = 'GPLv3+'

import logging
import os
from pathlib import Path
import re
import shutil
import subprocess
from zipfile import ZipFile

import numpy as np
from quicktions import Fraction

from viperleed.calc.lib.base import SingularMatrixError
from viperleed.calc.lib.base import cosvec
from viperleed.calc.lib.base import ensure_integer_matrix
from viperleed.calc.lib.base import lcm
from viperleed.guilib import get_equivalent_beams


# constants for conversion Angstrom and eV <-> atomic units
HARTREE_TO_EV = 27.211396
EV_TO_HARTREE = 1/HARTREE_TO_EV
BOHR_TO_ANGSTROM = 0.529177210903
ANGSTROM_TO_BOHR = 1/BOHR_TO_ANGSTROM

logger = logging.getLogger(__name__)


###############################################
#                FUNCTIONS                    #
###############################################

def get_superlattice_repetitions(matrix):
    """Return the number of repeats in 2D to cover a superlattice.

    Parameters
    ----------
    matrix : Sequence
        Shape (2, 2). The matrix representing the transformation
        between the "base" lattice and the super-lattice. The
        relation is suprlattice_basis = matrix @ base_basis,
        with the _basis matrices such that in-plane unit vectors
        are rows, i.e., a, b = _basis. Must be non-singular, and
        with all entries (close to) integers.

    Returns
    -------
    repeats : tuple
        Two elements, each corresponds to the number of repetitions
        of the "base" cell along its two unit vectors so as to
        completely cover the superlattice area when back-folded.

    Raises
    ------
    ValueError
        If matrix is singular or an inappropriate shape.
    NonIntegerMatrixError
        If matrix has non-integer elements.
    """
    # Work on a numpy-array copy of the input. Ensure it's float
    # to avoid UFuncTypeError when doing in-place divisions below
    matrix = ensure_integer_matrix(np.copy(matrix).astype(float))
    if matrix.shape != (2, 2):
        raise ValueError(f'Unexpected shape {matrix.shape} for superlattice '
                         'transform matrix. Should be (2, 2)')
    n_repeats = abs(matrix[0, 0] * matrix[1, 1] - matrix[1, 0] * matrix[0, 1])
    n_repeats = round(n_repeats)
    if not n_repeats:
        raise SingularMatrixError('Superlattice transform matrix is singular')

    if n_repeats == 1:  # No area change
        return 1, 1

    # The trick is making the input matrix into a lower-triangular
    # matrix by Gauss elimination. Then we read off the 'number of
    # repeats along the first base vector' from the only non-zero
    # element of the first row. The other direction is just the
    # ratio of n_repeats and the one we already know.
    if not matrix[1, 1]:  # The matrix is upper-triangular. Swap.
        matrix[[0, 1]] = matrix[[1, 0]]
    if matrix[0, 1]:      # The matrix is not triangular. Gauss.
        multiple = lcm(abs(round(matrix[0, 1])),  # lcm needs integers
                       abs(round(matrix[1, 1])))
        first_row, second_row = matrix
        first_row /= first_row[1]
        first_row -= second_row/second_row[1]
        first_row *= multiple  # Ensure integer
    return abs(round(matrix[0, 0])), abs(round(n_repeats / matrix[0, 0]))


def getYfunc(ivfunc, v0i):
    """
    Returns the Y function for a given function. The derivative is
    approximated by the slope between one point to the left and one to the
    right, so the Y function does not have values for the highest and lowest
    energies that are passed.
    Y = L/sqrt(1 + v0i^2 * L^2)
    L = ivfunc' / ivfunc

    Parameters
    ----------
    ivfunc : iterable
        Should be a 2D array, list of lists, or list of tuples containing
        energies in ivfunc[:,0] and values in ivfunc[:,1], both as float.

    v0i : float
        Imaginary part of the inner potential

    Returns
    -------
    yfunc : numpy array
        2D array of the Y-function energies and values

    """
    if len(ivfunc) < 3:
        # cannot calculate derivative, return empty
        return np.array([[]])
    yfunc = None
    ivfunc = np.array(ivfunc)
    for i in range(1, len(ivfunc)-1):
        vals = ivfunc[i-1:i+2, 1]
        lf = ((vals[2] - vals[0]) /
              (2*vals[1] + 1e-100))  # +1e-100 to avoid div by 0
        y = lf / (1 + (lf*v0i)**2)
        try:
            yfunc = np.append(yfunc, [[ivfunc[i, 0], y]], axis=0)
        except ValueError:
            if not yfunc:
                yfunc = np.array([[ivfunc[i, 0], y]])
            else:
                raise
    return yfunc


def getMaxTensorIndex(home=".", zip_only=False):
    """
    Checks the Tensors folder for the highest Tensor index there,
    returns that value, or zero if there is no Tensors folder or no valid
    Tensors zip file. zip_only looks only for zip files, ignoring directories.
    """
    tensor_dir = (Path(home) / "Tensors").resolve()
    if not tensor_dir.is_dir():
        return 0
    indlist = []
    rgx = re.compile(r'Tensors_[0-9]{3}\.zip')
    for f in [f for f in os.listdir(os.path.join(home, "Tensors"))
              if (os.path.isfile(os.path.join(home, "Tensors", f))
                  and rgx.match(f))]:
        m = rgx.match(f)
        if m.span()[1] == 15:  # exact match
            indlist.append(int(m.group(0)[-7:-4]))
    if not zip_only:
        rgx = re.compile(r'Tensors_[0-9]{3}')
        for f in [f for f in os.listdir(os.path.join(home, "Tensors"))
                  if ((tensor_dir / f).is_dir() and rgx.match(f))]:
            m = rgx.match(f)
            if m.span()[1] == 11:  # exact match
                indlist.append(int(m.group(0)[-3:]))
    if indlist:
        return max(indlist)
    return 0


def getDeltas(index, basedir=".", targetdir=".", required=True):
    """Fetches Delta files from Deltas or archive with specified tensor index.
    If required is set True, an error will be printed if no Delta files are
    found.
    basedir is the directory in which the Delta directory is based.
    targetdir is the directory to which the Tensor files should be moved."""
    dn = "Deltas_"+str(index).zfill(3)
    _basedir, _targetdir = Path(basedir).resolve(), Path(targetdir).resolve()
    zip_path=(_basedir / "Deltas" / dn).with_suffix(".zip")
    if os.path.isdir(_basedir / "Deltas" / dn):
        for f in [f for f in os.listdir(_basedir / "Deltas" / dn)
                  if (os.path.isfile(_basedir / "Deltas" / dn / f)
                      and f.startswith("DEL_"))]:
            try:
                shutil.copy2(_basedir / "Deltas" / dn / f, targetdir)
            except Exception:
                logger.error("Could not copy existing delta files to "
                             "work directory")
                raise
    elif os.path.isfile(zip_path):
        logger.info(f"Unpacking {dn}.zip...")
        try:
            with ZipFile(zip_path, 'r') as zip_ref:
                zip_ref.extractall(_targetdir)                                  # TODO: maybe it would be nicer to read directly from the zip file
        except Exception:
            logger.error(f"Failed to unpack {dn}.zip")
            raise
    elif required:
        logger.error("Deltas not found")
        raise RuntimeError("Deltas not found")
    return None


def fortran_compile_batch(tasks, retry=True, logname="fortran-compile.log"):
    """
    Performs a list of fortran compilations.

    Parameters
    ----------
    tasks : iterable of tuples (pre, filename, post)
        Each entry will be passed to fortran_compile.
    retry : bool, optional
        If compilation fails, check whether failure is likely due to missing
        '-mcmodel=medium', and if so, retry. The default is True.
    logname: str, optional
        Name of the log file.

    Returns
    -------
    None on success, else raises RuntimeException.

    """

    r = None
    for (pre, filename, post) in tasks:
        r = fortran_compile(pre=pre, filename=filename, post=post,
                            logname=logname)
        if r:
            break
    if not r:
        return None
    if r == "mcmodel":
        if not retry:
            raise RuntimeError(
                "Compiling fortran code failed, likely due to missing "
                "'-mcmodel=medium'")
        logger.warning("Compiling fortran code failed; retrying with "
                       "'-mcmodel=medium' option.")
        newtasks = [(pre + " -mcmodel=medium", filename, post)
                    for (pre, filename, post) in tasks]
        fortran_compile_batch(newtasks, retry=False, logname=logname)
    else:
        raise RuntimeError("Compiling fortran code failed, unknown return "
                           "value: " + r)
    return None


def fortran_compile(pre="", filename="", post="",
                    logname="fortran-compile.log"):
    """Assembles pre+filename+post to a filename, tries to execute via
    subprocess.run, raises an exception if it fails."""
    fc = pre+" "+filename+" "+post
    fcl = fc.split()
    sep = ""
    if os.path.isfile(logname):
        sep = "\n\n"
    try:
        with open(logname, "a") as log:
            log.write(sep + "############\n# COMPILING: " + fc
                      + "\n############\n\n")
        with open(logname, "a") as log:
            r = subprocess.run(fcl, stdout=log, stderr=log)
    except Exception:
        logger.error("Error compiling "+filename)
        raise
    if r.returncode not in (0, 1):
            raise RuntimeError("Fortran compiler subprocess returned "
                               f"{r.returncode}.")
    if r.returncode == 1 and "-mcmodel=medium" not in fc:
        mcmodel = False
        with open(logname, "r") as log:
            if "relocation truncated to fit:" in log.read():
                mcmodel = True
                logger.warning(f"Compiling file {filename} failed, likely due to "
                               "missing '-mcmodel=medium' flag.")
        if not mcmodel:
            raise RuntimeError("Fortran compiler subprocess returned "
                               f"{r.returncode}.")
        return "mcmodel"
    if r.returncode != 0:
            raise RuntimeError("Fortran compiler subprocess returned "
                               f"{r.returncode}.")
    return None


def checkLattice(ab, eps=1e-3):
    """Takes unit vectors a,b as a 2x2 matrix, returns (lat,t), where lat is
    a string "square", "rectangular", "hexagonal", "rhombic" or "oblique", and
    t is a 2x2 transformation matrix which will transform the cell to obtuse
    if it is rhombic or hexagonal."""
    # Author: Michele Riva; slightly modified for consistency by FK
    t = np.array([[1, 0], [0, 1]])
    c = cosvec(ab[0], ab[1])
    d = (np.linalg.norm(ab[0]) / np.linalg.norm(ab[1])) - 1
    if abs(c) < eps:  # angle is 90°
        if np.abs(d) < eps:
            lat = "square"
        else:
            lat = "rectangular"
    elif np.abs(d) < eps:  # rhombic or hex
        if c > eps:  # angle is acute -> redefine to make it obtuse
            t = np.dot(np.array([[0, -1], [1, 0]]), t)  # keeps the handedness
            ab = np.dot(t, ab)
        c = cosvec(ab[0], ab[1])
        if abs(c + 1/2) < eps:
            lat = "hexagonal"
        else:
            lat = "rhombic"
    else:
        lat = "oblique"
    return (lat, t)


def reduceUnitCell(ab, eps=1e-3):
    """Takes an obtuse unit cell as a (2x2) matrix and reduces it to minimum
    circumference, keeping the area constant. This might reduce oblique unit
    cells to rectangular or hexagonal ones. Returns (ab, t, celltype), where
    ab is the modified unit cell, t is the transformation matrix, and celltype
    is a string describing the unit cell ("square", "rectangular",
    "hexagonal", "rhombic" or "oblique")."""
    # Author: Michele Riva; slightly modified for consistency by FK
    (lat, t) = checkLattice(ab, eps=eps)
    ab = np.dot(t, ab)
    if lat == "oblique":
        # Transform lattice to have the shortest two vectors, with angle
        #  closest to 90°. This might bring it to rect, hex or rhombic.
        # If neither, will anyway transform to have the closest to rect.
        # ALGORITHM for reduction to closest to rect:
        # This is a discrete version of Gram-Schmidt's algorithm to find
        #  orthogonal bases
        # At each iteration:
        # - order vectors by norm, the shortest first
        # - determine the projection of the second on the first, and calculate
        #    the nearest integer kk
        # - subtract from the second the projection calculated above
        # - check whether now the second is the smallest. If yes, repeat,
        #    otherwise finished.
        swap = np.array([[1, 0], [0, 1]])
        # matrix that keeps track of whether a and b are swapped
        while True:  # Swap vectors if needed to get the shortest first
            if np.linalg.norm(ab[0]) > np.linalg.norm(ab[1]):
                t0 = np.array([[0, 1], [1, 0]])
            else:
                t0 = np.array([[1, 0], [0, 1]])
            swap = np.dot(t0, swap)
            t = np.dot(t0, t)
            ab = np.dot(t0, ab)
            kk = int(np.round(np.dot(ab[0], ab[1]) / np.dot(ab[0], ab[0])))
            t0 = np.array([[1, 0], [-kk, 1]])
            t = np.dot(t0, t)
            ab = np.dot(t0, ab)
            if np.linalg.norm(ab[0]) <= np.linalg.norm(ab[1]):
                break
        # Swap vectors back if they were overall swapped
        t = np.dot(swap, t)
        ab = np.dot(swap, ab)
#         END OF ALGORITHM. Now the lattice ab is closest to rectangular. It
#           might be still any shape (square, rect, hex, rhombic, oblique)
        lat, t0 = checkLattice(ab, eps=eps)
        t = np.dot(t0, t)
        ab = np.dot(t0, ab)
#       If ab is still oblique, try to see if it can be transformed to hex or
#         rhombic by choosing "a" not to be the shortest vector of all.
#       If possible, keep the new transformation. Otherwise, stick to the one
#         that makes it closest to rectangular.
        if lat == "oblique":  # lattice is still oblique
            # Re-swapping guarantees that that the matrix has on the first line
            #   the shortest possible vector, and on the second line the second
            #   shortest possible vector.
            # The only possible combinations that can lead to a rhombic/hex are
            #   a'=b+a or a'=b-a, depending on whether the angle is acute or
            #   obtuse, respectively
            t2 = swap
            ab = np.dot(swap, ab)

            c = cosvec(ab[0], ab[1])
            t0 = [[-int(np.sign(c)), 1], [0, 1]]
            t2 = np.dot(t0, t2)
            ab = np.dot(t0, ab)

            lat, t0 = checkLattice(ab, eps=eps)
            t2 = np.dot(t0, t2)

            if lat == "oblique":
                # lattice is still oblique, no transformation is needed (will
                #   keep the one closest to rect)
                t2 = np.array([[1, 0], [0, 1]])
        else:
            t2 = np.array([[1, 0], [0, 1]])
        t = np.dot(t2, t)
    return ab, t, lat


def reduce_c_vector(c_vec, ab_cell):
    """Compute a Minkowski-reduced version of the third unit-cell vector.

    Parameters
    ----------
    c_vec : numpy.ndarray
        The unit-cell vector to be reduced. Only the first
        two components are taken into consideration. They
        are modified in place.
    ab_cell : numpy.ndarray
        The in-plane unit cell to be used for reducing c.
        Unit vectors are rows, i.e., a, b == ab_cell.

    Returns
    -------
    None.
    """
    # The problem is essentially the closest-vector problem (CVP) that
    # is common in the theory of lattices: we have to find the lattice
    # vector (of the a,b lattice) that is closest to the projection
    # of c on the plane formed by ab_cell. That's the quantity to be
    # removed from c to make it shortest. The trick is that the ab_cell
    # should be Minkowski-reduced beforehand to ensure that also the c
    # vector will be Minkowski-reduced. The theory behind this can be
    # found in doi.org/10.1007/978-3-540-24847-7_26.
    ab_cell, *_ = reduceUnitCell(ab_cell)
    c_frac_ab = c_vec[:2].dot(np.linalg.inv(ab_cell))

    # The closest projection is an integer version of c_frac_ab.
    # It's easiest to take the floor, then consider also -a, -b,
    # and -(a+b), whichever gives the shortest vector. Otherwise
    # we'd have to also consider +a, +b, a-b, b-a, etc...
    c_frac_ab -= np.floor(c_frac_ab)

    increments = (0, 0), (1, 0), (0, 1), (1, 1)
    c_vec[:2] = min(
        ((c_frac_ab - f).dot(ab_cell) for f in increments),
        key=np.linalg.norm
        )


def bulk_3d_string(screws, glides):
    """Return info about bulk screw axes and glide planes as a string.

    Parameters
    ----------
    screws : Sequence
        Items are orders of rotation for screw axes.
    glides : Sequence
        Items are 2-tuples representing fractional coordinates
        of direction vectors of the glide planes.

    Returns
    -------
    bulk_3d_str : str
        Format is 'r(2, 4), m([1,1], [ 1,-1])' if there is
        any screw axes or glide planes, otherwise 'None'.
    """
    b3ds = ''
    if screws:
        screws_str = ', '.join(str(v) for v in screws)
        b3ds += f'r({screws_str})'
    if glides:
        if b3ds:
            b3ds += ', '
        glides_str = ', '.join(np.array2string(direction, separator=',')
                               for direction in glides)
        if glides_str:
            b3ds += f'm({glides_str})'
    return b3ds or 'None'


def getLEEDdict(sl, rp):
    """Return a LEED dict containing information needed by guilib functions."""
    if sl.planegroup == 'unknown':
        logger.warning('Generating LEED dictionary for slab with unknown '
                       'plane group!')
    if sl.planegroup in ['pm', 'pg', 'cm', 'rcm', 'pmg']:
        pgstring = sl.planegroup + '[{} {}]'.format(*sl.orisymplane.par)
    else:
        pgstring = sl.planegroup
    if not (abs(np.round(rp.SUPERLATTICE).astype(int) - rp.SUPERLATTICE)
            < 1e-3).all():
        logger.error('getLEEDdict: SUPERLATTICE contains non-integer-valued '
                     'entries.')
        return None
    # Some values can be overwritten via parameters:
    d = {'eMax': rp.THEO_ENERGIES.max,
         'SUPERLATTICE': rp.SUPERLATTICE.astype(int),
         'surfBasis': sl.ab_cell.T,
         'surfGroup': pgstring,
         'bulkGroup': sl.bulkslab.foundplanegroup,
         'bulk3Dsym': sl.bulkslab.get_bulk_3d_str(),
         'screenAperture': rp.SCREEN_APERTURE,
         'beamIncidence': (rp.THETA, rp.PHI)}
    # some values can be overwritten via parameters:
    if isinstance(rp.AVERAGE_BEAMS, tuple):
        d['beamIncidence'] = rp.AVERAGE_BEAMS
    if 'group' not in rp.SYMMETRY_BULK:
        return d

    # Some definitions for bulk symmetry.                                       # TODO: use guilib functions
    allowed_groups = {
        'oblique': ('p1', 'p2'),
        'rhombic': ('p1', 'p2', 'cm', 'cmm'),
        'rectangular': (
            'p1', 'p2', 'pm', 'pg', 'rcm', 'pmm', 'pmg', 'pgg', 'rcmm'),
        'square': (
            'p1', 'p2', 'pm', 'pg', 'cm', 'cmm', 'rcm', 'pmm', 'pmg', 'pgg',
            'rcmm', 'p4', 'p4m', 'p4g'),
        'hexagonal': (
            'p1', 'p2', 'cm', 'cmm', 'p3', 'p3m1', 'p31m', 'p6', 'p6m')
        }
    allowed_rotations = {
        'oblique': (2,),
        'rhombic': (2,),
        'rectangular': (2,),
        'square': (2, 4),
        'hexagonal': (2, 3, 4)
        }
    allowed_mirrors = {
        'oblique': (),
        'rhombic': ((1, 1), (1, -1)),
        'rectangular': ((1, 0), (0, 1)),
        'square': ((1, 0), (0, 1), (1, 1), (1, -1)),
        'hexagonal': ((1, 0), (0, 1), (1, 1), (1, -1), (1, 2), (2, 1))
        }
    if (rp.SYMMETRY_BULK['group'].split('[')[0]
            not in allowed_groups[sl.bulkslab.celltype]):
        hermann, *_ = rp.SYMMETRY_BULK['group'].split('[')
        logger.warning(
            f'Group {hermann} given in SYMMETRY_BULK is not allowed '
            f'for bulk cell shape {sl.bulkslab.celltype!r}.'
            )
        rp.setHaltingLevel(2)
    else:
        d['bulkGroup'] = rp.SYMMETRY_BULK['group']
    bulk_rotations = []
    bulk_mirrors = []
    if 'rotation' in rp.SYMMETRY_BULK:
        for order in rp.SYMMETRY_BULK['rotation']:
            if order not in allowed_rotations[sl.bulkslab.celltype]:
                logger.warning(
                    f'Rotation order {order} given in SYMMETRY_BULK is not '
                    f'allowed for bulk cell shape {sl.bulkslab.celltype!r}.'
                    )
                rp.setHaltingLevel(2)
            else:
                bulk_rotations.append(order)
    if 'mirror' in rp.SYMMETRY_BULK:
        for par in rp.SYMMETRY_BULK['mirror']:
            if par not in allowed_mirrors[sl.bulkslab.celltype]:
                logger.warning(
                    f'Mirror direction {par} given in SYMMETRY_BULK is not '
                    f'allowed for bulk cell shape {sl.bulkslab.celltype!r}.'
                    )
                rp.setHaltingLevel(2)
            else:
                bulk_mirrors.append(par)
    d['bulk3Dsym'] = bulk_3d_string(bulk_rotations, bulk_mirrors)
    return d


def getSymEqBeams(sl, rp):
    """Returns a list of tuples ((hf,kf), index), where (hf,kf) are beams and
    index is the group of other beams they are equivalent to"""
    if rp.AVERAGE_BEAMS is False:
        return []
    if not rp.domainParams:
        d = [getLEEDdict(sl, rp)]
    else:
        d = [getLEEDdict(dp.sl, dp.rp) for dp in rp.domainParams]
    if any([v is None for v in d]):
        logger.error("Failed to get beam equivalence list")
        return []
    symeqnames = get_equivalent_beams(*d)
    symeq = []
    rgx = re.compile(r'(?P<h>[-0-9/]+)\s*,\s*(?P<k>[-0-9/]+)')
    for (name, index) in symeqnames:
        m = rgx.match(name)
        if m:
            h, k = m.group("h"), m.group("k")
            try:
                hf, kf = float(Fraction(h)), float(Fraction(k))
            except ValueError:
                logger.error("getBeamCorrespondence: Could not convert beam "
                             "names from getEquivalentBeams to h,k floats")
                raise
            symeq.append(((hf, kf), index))
        else:
            logger.warning("getBeamCorrespondence: Beam name from "
                           "getEquivalentBeams not recognized: "+name)
    return symeq


def getBeamCorrespondence(sl, rp):
    """Compares theoretical and experimental beams, returns a list containing
    a number for each theoretical beam, indicating which experimental beam it
    corresponds to."""
    eps = 1e-5  # for comparing beams
    #   make dictionary {theoretical_beam: experimental_beam}
    beamcorr = {}
    # get symmetry-equivalence list:
    symeq = getSymEqBeams(sl, rp)
    # first, go through experimental beams and assign theoreticals if clear:
    remlist = []
    for eb in rp.expbeams:
        found = False
        for tb in rp.ivbeams:
            if eb.isEqual(tb, eps=eps):
                beamcorr[tb] = eb
                found = True
                break
        if not found:
            # check symmetry equivalent ones to exp beam
            eqbl = []   # hk of equivalent beams
            for (hk, i) in symeq:
                if eb.isEqual_hk(hk, eps=eps):
                    eqbl.extend([hk2 for (hk2, j) in symeq if i == j])
            for hk in eqbl:
                for tb in rp.ivbeams:
                    if tb.isEqual_hk(hk, eps=eps):
                        beamcorr[tb] = eb
                        found = True
                        break
                if found:
                    break
        if not found:
            logger.warning(
                "Experimental beam "+eb.label+" does not have a "
                "theoretical counterpart. Consider adding it to IVBEAMS.")
            rp.setHaltingLevel(1)
            remlist.append(eb)
    for b in remlist:
        rp.expbeams.remove(b)
        logger.warning("Beam "+b.label+" was removed from set of "
                       "experimental beams!")
    # NUMBER OF EXPERIMENTAL BEAMS MUST BE STATIC FROM HERE ON OUT
    # now, for theoretical beams without assignment, see if there is a
    #   symmetry-equivalent beam to assign them to
    notfound = []
    for tb in rp.ivbeams:
        if tb not in beamcorr.keys():
            found = False
            eqbl = []   # hk of equivalent beams
            for (hk, i) in symeq:
                if tb.isEqual_hk(hk, eps=eps):
                    eqbl.extend([hk2 for (hk2, j) in symeq if i == j])
            for hk in eqbl:
                for eb in rp.expbeams:
                    if eb.isEqual_hk(hk, eps=eps):
                        beamcorr[tb] = eb
                        found = True
                        break
                if found:
                    break
            if not found:
                notfound.append(tb.label)
    if notfound:
        logger.debug("No experimental beams found for calculated beams: "
                     + ", ".join(notfound))
    beamcorr_list = [-1]*len(rp.ivbeams)
    for i, tb in enumerate(rp.ivbeams):
        if tb in beamcorr:
            beamcorr_list[i] = rp.expbeams.index(beamcorr[tb])
    return beamcorr_list

# TODO: can eventually become part of compileTask class
def copy_compile_log(rp, logfile, log_name="fortran-compile"):
    """Copy compilation log file to compile_logs (will be moved to SUPP later).

    Parameters
    ----------
    rp : RunParameters
        rp object of the calculation.
    logfile : pathlike or str
        Path to the logfile that should be copied.
    log_name : str, optional
        Name to be used to identify logfile, eg. "refcalc". Default: "fortran-compile"
    """
    _logfile = Path(logfile)
    if rp.compile_logs_dir is None:
        # Compile directory not set. Cannot copy. Do not bother complaining.
        return
    target_path = (rp.compile_logs_dir / log_name).with_suffix(".log")
    try:
        shutil.copy2(_logfile, target_path)
    except OSError as err:
        logger.warning(
            f"Unable to copy compilation log file {str(_logfile)}. Info: {err}"
        )



"""Module errors of viperleed.calc.classes.slab.

Contains exceptions specific to BaseSlab objects.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-02-21'
__license__ = 'GPLv3+'

from viperleed.calc.classes.atom_containers import AtomContainerError


class SlabError(AtomContainerError):
    """Base exception for Slab objects."""


class AlreadyMinimalError(SlabError, RuntimeError):
    """A minimization was requested, but the quantity is already minimal."""


class AtomsTooCloseError(SlabError):
    """At least two atoms are too close, irrespective of their element."""


class EmptySlabError(SlabError):
    """The slab has no atoms."""


class InvalidUnitCellError(SlabError):
    """Exception raised when the unit cell of a slab is inappropriate."""


class MissingBulkSlabError(SlabError, RuntimeError):
    """A bulkslab attribute would be needed but it is not available."""


class MissingElementsError(SlabError):
    """The slab has no elements."""


class MissingLayersError(SlabError):
    """An operation requires layers defined."""


class MissingSublayersError(SlabError):
    """An operation requires sublayers defined."""


class NoBulkRepeatError(SlabError, RuntimeError):
    """Exception raised when failing to find a bulk repeat vector."""


class TooFewLayersError(SlabError):
    """Some layers are present, but it's too few for the operation."""


class VacuumError(SlabError):
    """Something is wrong with the vacuum gap of a slab."""

    def __init__(self, message, fixed_slab):
        """Initialize exception instance.

        Parameters
        ----------
        message : str
            Textual message to be displayed for this exception
        fixed_slab : SurfaceSlab
            A slab that has been modified to fix the problem with
            the vacuum gap.

        Returns
        -------
        None.
        """
        self.message = message
        self.fixed_slab = fixed_slab
        super().__init__(message)


class NotEnoughVacuumError(VacuumError):
    """A slab has too little vacuum."""


class NoVacuumError(NotEnoughVacuumError):
    """A slab has no vacuum gap at all."""


class WrongVacuumPositionError(VacuumError):
    """A slab has enough vacuum, but it's somewhere in the middle."""



"""Functions for writing output from full-dynamic optimization."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2021-10-25'
__license__ = 'GPLv3+'

import copy
import logging

import numpy as np
from numpy.polynomial import Polynomial

try:
    import matplotlib
    matplotlib.rcParams.update({'figure.max_open_warning': 0})
    matplotlib.use('Agg')  # !!! check with Michele if this causes conflicts
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt
    plt.style.use('viperleed.calc')
    from matplotlib import cm
except Exception:
    _CAN_PLOT = False
else:
    _CAN_PLOT = True

from viperleed.calc.files.iorfactor import read_rfactor_columns
from viperleed.calc.files.ivplot import plot_iv


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def write_fd_opt_csv(points, which, filename="FD_Optimization.csv", sep=","):
    """
    Writes results from the full dynamic optimization into a CSV file

    Parameters
    ----------
    points : numpy.array
        Two columns, containing points and corresponding R-factors
    which : str
        The parameter that was being optimized, for column title
    filename : str, optional
        Name of the csv file to write. The default is "FD_Optimization.csv".
    sep : str, optional
        The separator to use. The default is ",".

    Returns
    -------
    None.

    """

    sorted_points = points[points[:, 0].argsort()]
    title = which
    if which in ["a", "b", "c", "ab", "abc"]:
        title += " scaling"
    width = max(len(title), 12)

    output = title.rjust(width) + sep + "R".rjust(width) + "\n"
    for xy in sorted_points:
        output += "{:.5f}".format(xy[0]).rjust(width) + sep
        output += "{:.5f}".format(xy[1]).rjust(width) + "\n"

    try:
        with open(filename, "w") as wf:
            wf.write(output)
    except Exception as e:
        logger.warning("Failed to write "+filename + ": " + str(e))
    return


def write_fd_opt_pdf(points, which, filename="FD_Optimization.pdf",
                     parabola=None):
    """
    Plots results from the full dynamic optimization into a pdf file

    Parameters
    ----------
    points : numpy.array
        Two columns, containing points and corresponding R-factors
    which : str
        The parameter that was being optimized, for x-axis label
    filename : str, optional
        File name to write to. The default is "FD_Optimization.pdf".
    parabola : Polynomial, optional
        Pre-existing fit parabola. If not passed, data will be fit again here.

    Returns
    -------
    None.

    """

    global _CAN_PLOT
    if not _CAN_PLOT:
        logger.debug("Necessary modules for plotting not found. Skipping "
                     "error plotting.")
        return
    title = which
    if which in ["a", "b", "c", "ab", "abc"]:
        title += " scaling"

    fig = plt.figure(figsize=(5.8, 4.1))
    ax = fig.add_subplot(1, 1, 1)
    ax.set_xlabel(title)
    ax.set_ylabel('Pendry R-factor')

    if len(points) > 2:
        if parabola is None:
            parabola = Polynomial.fit(points[:, 0], points[:, 1], 2)
        coefs = parabola.convert(domain=[-1, 1]).coef
        p_min = -0.5*coefs[1] / coefs[2]
        p_r = parabola(p_min)
        xx, yy = parabola.linspace()
        ax.plot(xx, yy, lw=2)
        xlims = (min(points[:, 0]), max(points[:, 0]))   # data range x
        ylims = (min(points[:, 1]), max(points[:, 1]))   # data range y
        namePos = ((xlims[1] + xlims[0])*0.5,
                   ylims[1] - 0.1*(ylims[1] - ylims[0]))
        if coefs[2] > 0:
            ax.annotate("Minimum at {:.4f}\nR = {:.4f}".format(p_min, p_r),
                        namePos, fontsize=10, ha="center")
    ax.plot(points[:, 0], points[:, 1], 'o', c='darkslategray')
    fig.tight_layout()

    # write
    try:
        pdf = PdfPages(filename)
        pdf.savefig(fig)
    except PermissionError:
        logger.warning("Failed to write to " + filename
                       + ": Permission denied.")
    except KeyboardInterrupt:
        raise
    except Exception:
        logger.warning("Failed to write to "+filename, exc_info=True)
    finally:
        try:
            pdf.close()
        except Exception:
            pass
    try:
        plt.close(fig)
    except Exception:
        pass
    return


def write_fd_opt_beams_pdf(rp, points, which, tmpdirs, best_rfactors,
                           filename="FD_Optimization_beams.pdf"):
    """
    Plots all I(V) curves that were calculated during the optimization.

    Parameters
    ----------
    rp : Rparams
        Run parameters.
    points : numpy.array
        Two columns, containing points and corresponding R-factors.
    which : str
        The parameter that was being optimized, for legend.
    tmpdirs : list of str
        Paths of directories from which I(V) data should be read.
    best_rfactors : list of float
        The R-factors from the best calculation. Will be written into plots.
    filename : str, optional
        File name to write to. The default is "FD_Optimization_beams.pdf".

    Returns
    -------
    None.

    """

    global _CAN_PLOT
    if not _CAN_PLOT:
        logger.debug("Necessary modules for plotting not found. Skipping "
                     "error plotting.")
        return

    new_order = points[:, 0].argsort()
    tmpdirs = list(np.array(tmpdirs)[new_order])
    points = points[new_order]
    best_point_ind = np.argmin(points[:, 1])

    theodata = []
    exp_to_use = None

    for i, path in enumerate(tmpdirs):
        try:
            theospec, expspec = read_rfactor_columns(path)
        except Exception:
            logger.warning("Error reading calculated spectra for point {:.4f}"
                           .format(points[i, 0]))
            continue
        theodata.append(theospec)
        if i == best_point_ind:
            exp_to_use = expspec
    if exp_to_use is None:
        logger.warning("Failed to read in experimental spectra from best "
                       "full-dynamic run. Output of collected I(V) spectra "
                       "will be skipped.")
        return
    legends = ["{} = {:.4f}".format(which, points[i, 0])
               for i in range(len(points))] + ["Experimental"]
    annotations = []
    if best_rfactors:
        annotations = ["R = {:.4f}".format(r) for r in best_rfactors]
    if rp.PLOT_IV["overbar"]:
        labelstyle = "overbar"
    else:
        labelstyle = "minus"
    labelwidth = max([beam.getLabel(style=labelstyle)[1]
                      for beam in rp.expbeams])
    labels = [b.getLabel(lwidth=labelwidth, style=labelstyle)[0]
              for b in rp.expbeams]
    formatting = copy.deepcopy(rp.PLOT_IV)  # use to set colors
    formatting['colors'] = (
        list(cm.get_cmap('viridis', len(points)).colors)
        + [np.array([0, 0, 0, 1])])
    formatting['curve_line_widths'] = [0.5] * len(points) + [1.]
    formatting['curve_line_widths'][best_point_ind] = 1.
    try:
        plot_iv(theodata + [exp_to_use], filename,
                labels=labels, annotations=annotations, legends=legends,
                formatting=formatting)
    except Exception:
        logger.warning("Error plotting collected I(V) curves.",
                       exc_info=True)
    return



"""Module layer_cuts of viperleed.calc.classes.rparams.special.

Defines the LayerCutTokenType, LayerCutToken, and LayerCuts classes.
They are convenience classes for handling user input of slab cut
positions used to generate layers of a Slab. The code is a rewrite
of the functionality, originally by @fkraushofer, previously split
partly in ParameterInterpreter, partly in Slab.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-21'
__license__ = 'GPLv3+'

from collections.abc import Sequence
from dataclasses import dataclass
from enum import IntEnum, auto
import itertools
from numbers import Real
import re
from typing import Any

from viperleed.calc.lib.base import pairwise

from .base import SpecialParameter


def threewise(iterable):
    """Yield triplets of items from iterable."""
    orig, shift_one, shift_two = itertools.tee(iterable, 3)
    next(shift_one, None)
    next(shift_two, None)
    next(shift_two, None)
    yield from zip(orig, shift_one, shift_two)


class LayerCutTokenType(IntEnum):
    """Possible types of layer cut tokens."""

    NUMERIC = auto()
    ORDERING = auto()  # < or >
    AUTO_DC = auto()
    AUTO_DZ = auto()
    INVALID = auto()


_AUTO_CUT_RE = re.compile(r'(?P<type>dz|dc)\((?P<cutoff>[\d.]+)\)')


@dataclass(frozen=True)
class LayerCutToken:
    """A single token of a LayerCuts object."""

    type_: LayerCutTokenType
    value: Any

    # Two attributes that can only be set once, and only for auto-cut
    lower: 'LayerCutToken' = None
    upper: 'LayerCutToken' = None

    def __eq__(self, other):
        """Return whether this LayerCutToken is equal to other."""
        if isinstance(other, str):
            other = LayerCutToken.from_string(other)
        elif isinstance(other, Real):
            other = LayerCutToken.make_numeric(other)
        if not isinstance(other, LayerCutToken):
            return NotImplemented
        equal = (self.type_ is other.type_ and self.value == other.value
                 and self.lower == other.lower and self.upper == other.upper)
        return equal or NotImplemented

    def __format__(self, format_spec):
        """Return a formatted version of this LayerCutToken.

        Parameters
        ----------
        format_spec : str
            The formatting specification for this token. Currently
            format_spec is used only if token.is_numeric. It is
            then passed unchanged to self.value.__format__, i.e.,
            to float.__format__.

        Returns
        -------
        formatted : str
            Formatted version of this token.
        """
        if not self.is_numeric:
            return str(self)
        return format(self.value, format_spec)

    def __str__(self):
        """Return a string version of this token."""
        if self.is_ordering:
            return self.value
        if self.is_numeric:
            return str(self.value)
        if self.is_auto_cut:
            dc_dz = self.type_.name.replace('AUTO_', '').lower()
            return f'{dc_dz}({self.value})'
        return 'INVALID'

    def __post_init__(self):
        """Check that the initialization parameters are acceptable."""
        if ((self.is_numeric or self.is_auto_cut)
                and not isinstance(self.value, Real)):
            raise TypeError('Numeric tokens must have numeric values')
        if self.is_numeric and not 0 <= self.value <= 1:
            raise ValueError(f'Numeric token value {self.value} '
                             'is not in range [0, 1]')
        if self.is_ordering and self.value not in '<>':
            raise ValueError('Ordering tokens can only have < or > as value')
        if not self.is_auto_cut and (self.lower, self.upper) != (None, None):
            raise ValueError('Only dc/dz can have bounds')
        if isinstance(self.lower, Real):
            object.__setattr__(self, 'lower', self.make_numeric(self.lower))
        if isinstance(self.upper, Real):
            object.__setattr__(self, 'upper', self.make_numeric(self.upper))
        if self.is_auto_cut:
            self._check_valid_bound_type(self.lower, self.upper)

    def __repr__(self):
        """Return a representation string for this LayerCutToken."""
        repr_ = f'LayerCutToken({self.type_.name}, {self.value}'
        if self.is_auto_cut and (self.lower, self.upper) != (None, None):
            repr_ += f', {self.lower}, {self.upper}'
        return repr_ + ')'

    @property
    def is_auto_cut(self):
        """Return whether this token is of type 'dc'/'dz'."""
        return self.type_ in (LayerCutTokenType.AUTO_DC,
                              LayerCutTokenType.AUTO_DZ)

    @property
    def is_auto_cut_dc(self):
        """Return whether this token is of type 'dc'."""
        return self.type_ is LayerCutTokenType.AUTO_DC

    @property
    def is_ordering(self):
        """Return whether this token is '>' or '<'."""
        return self.type_ is LayerCutTokenType.ORDERING

    @property
    def is_larger_than(self):
        """Return whether this token is a '>'."""
        return self.is_ordering and self.value == '>'

    @property
    def is_numeric(self):
        """Return whether this token is a number."""
        return self.type_ is LayerCutTokenType.NUMERIC

    @classmethod
    def from_string(cls, string):
        """Return a LayerCutToken from a `string`."""
        string = re.sub(r'\s', '', string).lower()
        _invalid = cls(LayerCutTokenType.INVALID, string)
        if len(string) == 1 and string in '<>':
            return cls(LayerCutTokenType.ORDERING, string)
        try:
            value = float(string)
        except ValueError:
            pass
        else:
            try:
                return cls(LayerCutTokenType.NUMERIC, value)
            except ValueError:
                return _invalid
        auto_match = _AUTO_CUT_RE.match(string)
        if not auto_match:
            return _invalid
        try:
            value = float(auto_match['cutoff'])
        except ValueError:
            return _invalid
        type_name = f'AUTO_{auto_match["type"].upper()}'
        return cls(LayerCutTokenType[type_name], value)

    @classmethod
    def make_numeric(cls, value):
        """Return a LayerCutToken of type NUMERIC from a numeric `value`."""
        return cls(LayerCutTokenType.NUMERIC, float(value))

    def get_bounds(self, other_lower_bounds=()):
        """Return the numeric value of lower and upper bounds.

        This method assumes that this token is used internally in
        a LayerCuts object. This means that the scope of this token
        is bound only with '<', and not with '>'.

        Parameters
        ----------
        other_lower_bounds : Sequence, optional
            Other bounds to be considered for the lower side.
            Items are numbers. Default is no extra bounds.

        Raises
        ------
        AttributeError
            If this method is called on a non-auto-cut LayerCutToken
        RuntimeError
            If this token has its lower bound larger than its higher
            bound.
        ValueError
            If `other_lower_bounds` contains values larger than the
            upper bound of this token.
        """
        if not self.is_auto_cut:
            raise AttributeError(f'{self.type_.name} has no bounds')

        self_lower = self.lower.value if self.lower is not None else 0.0
        self_upper = self.upper.value if self.upper is not None else 1.0
        if self_lower >= self_upper:
            raise RuntimeError(f'{self} has swaped upper and lower bounds')
        if any(b > self_upper for b in other_lower_bounds):
            raise ValueError(
                'All other_lower_bounds must be at most as large '
                f'as the upper bound ({self_upper}) of this token'
                )
        return max((self_lower, *other_lower_bounds)), self_upper

    def with_bounds(self, lower, upper):
        """Return a copy of self with `lower` and `upper` bounds."""
        if not self.is_auto_cut:
            raise TypeError('Only auto-cut can have bounds')
        if any(attr is not None for attr in (self.lower, self.upper)):
            raise RuntimeError(f'{self} already has bounds')
        if isinstance(lower, Real):
            lower = self.make_numeric(lower)
        if isinstance(upper, Real):
            upper = self.make_numeric(upper)
        self._check_valid_bound_type(lower, upper)
        cls = type(self)
        return cls(self.type_, self.value, lower, upper)

    @staticmethod
    def _check_valid_bound_type(*bounds):
        """Ensure bounds are NUMERIC LayerCutToken or None."""
        try:
            all_numbers = all((b is None or b.is_numeric) for b in bounds)
        except AttributeError:  # Not a LayerCutToken
            raise TypeError('Bounds must be LayerCutToken objects') from None
        if not all_numbers:
            raise TypeError('Only LayerCutTokenType.NUMERIC-type '
                            'tokens can be used as bounds')


class LayerCuts(SpecialParameter, param='LAYER_CUTS'):
    """A container of LayerCutToken objects."""

    def __init__(self, *tokens):
        """Initialize instance from some LayerCutToken objects."""
        self._tokens = None
        self._tokens_with_ordering = None   # Also, sorted as original
        self._had_larger_than = False  # Internally we change all to <

        self._check_tokens(tokens)
        ori_tokens, tokens = self._clean_up_tokens(tokens)
        self._tokens = tokens
        self._tokens_with_ordering = ori_tokens

    def __bool__(self):
        """Return whether there is any cut in this LayerCuts."""
        return bool(self._tokens)

    def __format__(self, format_spec):
        """Return a formatted version of this LayerCuts object.

        Parameters
        ----------
        format_spec : str
            Format specification. This is passed unchanged to
            each of the tokens in this LayerCuts. See also
            help(LayerCutToken.__format__).

        Returns
        -------
        formatted : str
            Formatted version of this LayerCuts.
        """
        return ' '.join(format(t, format_spec)
                        for t in self._tokens_with_ordering)

    def __iter__(self):
        """Return an iterator of tokens in this LayerCuts object."""
        return iter(self._tokens)

    def __len__(self):
        """Return the number of cut specifications (excluding '<'/'>')."""
        return len(self._tokens)

    def __str__(self):
        """Return a string version of this LayerCuts object."""
        return ' '.join(str(t) for t in self._tokens_with_ordering)

    def __repr__(self):
        """Return a string representation of this LayerCuts."""
        return f'LayerCuts({str(self)!r})'

    @classmethod
    def as_layer_cuts(cls, other):
        """Return a LayerCuts from other, raise otherwise."""
        if isinstance(other, LayerCuts):
            return other
        if isinstance(other, str):
            return cls.from_string(other)     # May raise ValueError
        if isinstance(other, Sequence):
            cuts = cls()
            cuts.update_from_sequence(other)  # May Value/TypeError
            return cuts
        raise TypeError('Cannot produce a LayerCuts from '
                        f'{type(other).__name__!r}. Only '
                        'str and Sequence allowed')

    # Override parent's method
    from_value = as_layer_cuts

    @classmethod
    def from_string(cls, string):
        """Return LayerCuts from a `string`."""
        if all(c in string for c in '<>'):
            raise ValueError('Cannot parse list with both "<" and ">"')
        tokens = cls._tokenize_string(string)
        instance = cls(*tokens)
        return instance

    def update_from_sequence(self, sequence):
        """Clear this LayerCuts, and fill it with values from `sequence`.

        Parameters
        ----------
        sequence : Sequence
            Items must be LayerCutToken objects, real-valued numbers or
            tokenized strings. Each item is used to create the correct
            LayerCutToken.
        """
        new_tokens = []
        for item in sequence:
            if isinstance(item, LayerCutToken):
                token = item
            elif isinstance(item, Real):
                token = LayerCutToken.make_numeric(item)
            elif isinstance(item, str):
                token = LayerCutToken.from_string(item)
            else:
                raise TypeError('Expected Real or str, '
                                f'found {type(item).__name__}')
            new_tokens.append(token)
        self._check_tokens(new_tokens)
        ori_tokens, tokens = self._clean_up_tokens(new_tokens)
        self._tokens = tokens
        self._tokens_with_ordering = ori_tokens

    @staticmethod
    def _check_consistency_of_auto_cuts(tokens):
        """Raise if auto-cut tokens are inconsistent.

        Parameters
        ----------
        tokens : Sequence
            The tokens to be checked.

        Raises
        ------
        ValueError
            If auto-cut tokens do not have an ORDERING operator next
            to them (except when there's only one auto-cut item).
        """
        has_auto_cut = any(t.is_auto_cut for t in tokens)
        if not has_auto_cut or len(tokens) == 1:
            return
        err = 'dc/dz always requires < or > unless it is the only cut'
        for pair in pairwise(tokens):
            if not any(tok.is_auto_cut for tok in pair):
                continue
            left, right = pair
            if left.is_auto_cut and right.is_ordering:
                continue
            if left.is_ordering and right.is_auto_cut:
                continue
            raise ValueError(err)

    @staticmethod
    def _check_consistency_of_bounds(tokens):
        """Complain if any of the bounds of `tokens` is inconsistent.

        Parameters
        ----------
        tokens : Sequence
            The LayerCutToken objects to be checked. They are assumed
            to be already sorted so that only '<' bounds are used.

        Raises
        ------
        ValueError
            If any of the auto-cuts has its lower bound larger than
            its upper one
        ValueError
            If any fractional cuts falls within an 'auto-cut' range
        """
        auto_cuts = [t for t in  tokens if t.is_auto_cut]
        if not auto_cuts:
            return

        # First check that all the dc/dz have lower <= upper. While
        # going through, also collect the bounds for later checks.
        bounds = []
        for token in auto_cuts:
            try:
                bounds.append(token.get_bounds())
            except RuntimeError:
                raise ValueError(f'{token} has swapped lower '
                                 'and upper bounds') from None

        # Now check that none of the numeric cuts falls within
        # any of the auto-cut bounds
        for token in tokens:
            if not token.is_numeric:
                continue
            try:
                invalid = next(i for i, (lower, upper) in enumerate(bounds)
                               if lower < token.value < upper)
            except StopIteration:
                continue
            raise ValueError(f'Numeric cut {token.value} falls within '
                             f'the bounds of {auto_cuts[invalid]}')

    @staticmethod
    def _check_consistency_of_order_operators(tokens):
        """Raise if ordering operators are inconsistent.

        This method filters out weird input like the following:
        - '< dc [...]' and '[...] dc <'
        - 'N N < < dc'
        - 'N < N' and 'dc < dc'

        Parameters
        ----------
        tokens : Sequence
            Elements are LayerCutToken objects.

        Raises
        ------
        ValueError
            If an ORDERING token appears as the first or last
            item in tokens.
        ValueError
            If an ORDERING token does not separate a NUMERIC
            from an AUTO token.
        """
        has_ordering = any(t.is_ordering for t in tokens)
        if not has_ordering:
            return

        # Complain if '<' is first or last. Notice that this also
        # implies that, if this test passes, there are at least
        # three items.
        if any(tokens[i].is_ordering for i in (0, -1)):
            raise ValueError('Cannot have < or > as first or last items')

        # Check that left and right of a '<' or '>' there's
        # a number on one side and a dc/dz on the other
        for left, middle, right in threewise(tokens):
            if not middle.is_ordering:
                continue
            if left.is_numeric and right.is_auto_cut:
                continue
            if right.is_numeric and left.is_auto_cut:
                continue
            invalid = f'{left.value} {middle.value} {right.value}'
            raise ValueError(f'Invalid token triplet {invalid!r}. > / < '
                             'must separate a number and a dc / dz')

    @classmethod
    def _check_tokens(cls, tokens):
        """Run some basic consistency checks on `tokens`."""
        if any(not isinstance(t, LayerCutToken) for t in tokens):
            raise TypeError('Can only contain LayerCutToken items')

        invalid = [repr(t.value) for t in tokens
                   if t.type_ is LayerCutTokenType.INVALID]
        if invalid:
            raise ValueError('Could not parse ' + ', '.join(invalid))
        # Make sure that ordering operators are OK, i.e., they do not
        # appear first and last, and they separate a number and dc/dz
        cls._check_consistency_of_order_operators(tokens)
        cls._check_consistency_of_auto_cuts(tokens)

    def _clean_up_tokens(self, tokens):
        """Return tokens and the filtered copy used internally."""
        # We assume that tokens have already gone through basic
        # checks, i.e., they passed _check_tokens. Store a copy
        # to return before messing with tokens.
        ori_tokens = list(tokens)

        # Turn the tokens around if there is a '>'. This allows
        # us not to store the ordering operators at all (except
        # for the purpose of returning a string).
        self._had_larger_than = any(t.is_larger_than for t in tokens)
        if self._had_larger_than:
            tokens = reversed(tokens)

        # Now get rid of all the '<'/'>'
        tokens = [t for t in tokens if not t.is_ordering]

        # Assign 'bounds' to the dc/dz items. This also filters
        # out weird input like 'dz dc < N'. It is important to
        # realize that tokens is modified in-place and its items
        # are replaced by new ones. This makes the ori_tokens
        # above and the ones in tokens different. Should we ever
        # need to do anything other than str on the ori_tokens,
        # we need to deal with this appropriately here.
        self._set_bounds_of_auto_tokens(tokens)

        # And check (i) lower <= upper, (ii) no numbers fall in bounds
        self._check_consistency_of_bounds(tokens)
        return ori_tokens, tokens

    @staticmethod
    def _set_bounds_of_auto_tokens(tokens):
        """Store bounds information in the dc/dz items of `tokens`."""
        if len(tokens) <= 1:
            return
        shift_left, shift_right = (None, *tokens[:-1]), (*tokens[1:], None)
        triplets = zip(shift_left, tokens, shift_right)
        for i, (lower, middle, upper) in enumerate(triplets):
            if not middle.is_auto_cut:
                continue
            tokens[i] = middle.with_bounds(lower, upper)

    @staticmethod
    def _tokenize_string(string):
        """Return a list of LayerCutToken from a `string`."""
        # Make sure ">"/"<" are split off from the rest
        string = re.sub(r'([<>])', r' \1 ', string.strip()).lower()
        return [LayerCutToken.from_string(s) for s in string.split()]



"""ViPErLEED utility to convert experimental input for SATLEED to EXPBEAMS.csv.

Reads a file containing experimental I(V) curves as used by SATLEED
and converts it into the standard CSV format used by ViPErLEED.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-20'
__license__ = 'GPLv3+'

import logging
from pathlib import Path

import fortranformat as ff
import numpy as np

from viperleed.calc.classes.beam import Beam
from viperleed.calc.files.beams import averageBeams
from viperleed.calc.files.beams import writeOUTBEAMS
from viperleed.cli_base import ViPErLEEDCLI
from viperleed.guilib.base import BeamIndex
from viperleed.utilities.beams import EXPBEAMS_DEFAULT

logger = logging.getLogger(__name__)


_HELP = ('convert a SATLEED-style experimental-'
         'beams input file to ViPErLEED format')
_SCALING_FACTOR_DEVIATION_FROM_ONE = 1e-3  # Warn if > than this


class FromSATLEEDCLI(ViPErLEEDCLI, cli_name='from-SATLEED', help_=_HELP):
    """Utility to convert SATLEED beams file to EXPBEAMS format."""

    def add_parser_arguments(self, parser):
        """Add command-line arguments to parser."""
        super().add_parser_arguments(parser)
        parser.add_argument(
            'input',
            help='input file containing the I(V) curves as used by SATLEED',
            type=str,
            )
        parser.add_argument(
            '-o', '--output',
            help=('output file to write the beams to. '
                  f'Default is {EXPBEAMS_DEFAULT}'),
            type=str,
            default=EXPBEAMS_DEFAULT,
            )
        parser.add_argument(
            '--average',
            help=('Average beams according to the averaging '
                  'scheme in the input file. Default is False.'),
            action='store_true',
            default=False,
            )

    def __call__(self, args=None):
        """Parse CLI arguments, then convert file."""
        args = self.parse_cli_args(args)
        # Read input file
        try:
            iv_beams_data = _read_satleed_file(Path(args.input),
                                               average=args.average)
        except ValueError as exc:
            self.parser.error(str(exc))
            raise  # Unreachable as .error does sys.exit
        except StopIteration:
            self.parser.error(f'Not enough lines in {args.input}.')
            raise  # Unreachable as .error does sys.exit
        # Check if out_file exists...
        out_file = args.output
        if Path(out_file).exists():
            err_ = f'Output file {out_file} already exists.'
            self.parser.error(err_)
            raise FileExistsError(err_)
        # ...and write to it
        writeOUTBEAMS(filename=out_file, beams=iv_beams_data)


def _average_beams_if_requested(beams, averaging_scheme, average):
    """Return a list of averaged beams.

    Parameters
    ----------
    beams : list of Beam
        The beams to be averaged.
    averaging_scheme : Sequence of int
        Grouping indices for averaging `beams`. Same length as
        `beams`. Elements are group 'ids'. Beams with the same
        group id are averaged together.
    average: bool
        Whether any averaging should be carried out.

    Returns
    -------
    averaged_beams : list of Beam
        The beams averaged according to `averaging_scheme`.

    Raises
    ------
    ValueError
        If no averaging was requested via `average`, but the
        `averaging_scheme` would suggest averaging is needed.
    """
    averaging_groups = np.unique(averaging_scheme)
    if not average and averaging_groups.size != len(beams):
        raise ValueError(
            'Averaging scheme contains duplicate values. '
            'Use the --average option to average the beams if '
            'this is intended.'
            )
    if not average:
        return beams

    averaged_beams = []
    for group_id in averaging_groups:
        group_beams = [
            [beams[i]]
            for i in np.argwhere(averaging_scheme == group_id).flatten()        # TODO: probably nicer to do this with a boolean mask after converting beams into an array. Something along the lines of group_beams = beams[averaging_scheme == group_id]
            ]
        try:
            averaged_beams.append(averageBeams(group_beams, weights=None))
        except ValueError as exc:
            logger.warning(f'Failed to average group {group_id}: {exc}')
    return averaged_beams


def _read_averaging_scheme(line, n_beams):
    """Return a list of indices for beams to be averaged."""
    averaging_scheme = [int(x) for x in line.split()]
    if len(averaging_scheme) != n_beams:
        raise ValueError('Averaging scheme (line 3) has length '
                         f'{len(averaging_scheme)}, expected {n_beams}.')
    return averaging_scheme


def _read_beams(lines, n_beams, energy_increment, reader):
    """Return a list of `n_beams` Beam objects read from lines."""
    beam_data = {}
    for _ in range(n_beams):
        # read beam hk
        beam_id = next(lines).strip()
        # strip brackets left and right an convert to BeamIndex
        beam_id = beam_id[1:-1] if beam_id.startswith('(') else beam_id
        beam_id = BeamIndex(beam_id)

        # number of energies and scaling factor
        n_energies, scaling_factor = next(lines).split()
        n_energies = int(n_energies)
        scaling_factor = float(scaling_factor)

        if abs(scaling_factor - 1.0) > _SCALING_FACTOR_DEVIATION_FROM_ONE:
            logger.warning(
                f'Scaling factor for beam {beam_id} is {scaling_factor}.'
                )

        block_lines = (next(lines) for _ in range(n_energies))
        block_values = [reader.read(line) for line in block_lines]
        energies = np.array([value[0] for value in block_values])
        intensities = np.array([value[1] for value in block_values])

        # check that energies are equally spaced
        # this also checks that the energies are sorted
        if not np.allclose(np.diff(energies), energy_increment):
            raise ValueError(
                f'Energies for beam {beam_id} are not spaced '
                'according to the energy increment specified on '
                'line 5. Interpolation is not supported.'
                )

        # store into the dict and process the next beam
        beam_data[beam_id] = (energies, intensities)

    processed_beams = []
    for beam_id, (energies, intensities) in beam_data.items():
        hk_beam = Beam(beam_id)
        hk_beam.intens = dict(zip(energies, intensities))
        processed_beams.append(hk_beam)
    return processed_beams


def _read_satleed_file(file_name, average=False):
    """Read `file_name` and return the contents as a list of Beam objects.

    Parameters
    ----------
    file_name : Path
        Path to the input file.
    average : bool, optional
        If True, average beams according to the averaging
        scheme in `file_name`. Default is False.

    Returns
    -------
    list of Beam
        List of Beam objects containing the data from the input file.

    Raises
    ------
    ValueError
        If the file is not formatted correctly, or if
        averaging would be needed but `average` was False.
    StopIteration
        If there are not enough lines in `file_name`.
    """
    with file_name.open('r', encoding='utf-8') as file:
        lines = iter(file.readlines())

    # First line is name of the system
    _ = next(lines).strip()
    # number of beams
    n_beams = int(next(lines))

    # Averaging scheme
    averaging_scheme = _read_averaging_scheme(next(lines), n_beams)

    # Formatting
    data_format = next(lines).strip()
    reader = ff.FortranRecordReader(data_format)

    # Energy increment
    energy_increment = float(next(lines).strip())

    # Beam energies and intensities for each beam index
    processed_beams = _read_beams(lines, n_beams, energy_increment, reader)

    # Check that we have read all lines
    for surplus_line in lines:
        if surplus_line.strip():
            raise ValueError('Found surplus data after reading all beams.')

    # Average if requested
    return _average_beams_if_requested(processed_beams,
                                       averaging_scheme,
                                       average)


if __name__ == '__main__':
    FromSATLEEDCLI.run_as_script()



"""
======================================
  ViPErLEED Graphical User Interface
======================================
   *** Module guilib.decorators ***

Author: Michele Riva
Created: 2020-01-29

This module provides Qt-independent decorators based on the wrapt Python module
"""
import inspect
import cProfile
import pstats
from time import perf_counter as timer

import wrapt
try:
    from line_profiler import LineProfiler
except ImportError:
    pass


def ensure_decorates_class(superclass=object):
    """
    Decorator to be applied to a @wrapt.decorator. Raises exceptions if the
    @wrapt.decorator on which it acts is not applied to a subclass of
    superclass.

    Parameters
    ----------
    superclass: class
                default: object

    Returns
    -------
    The original @wrapt.decorator if the @wrapt.decorator is applied to a
    subclass of superclass. Raises RuntimeError otherwise.
    """

    @wrapt.decorator
    def _wrapper(decorator, always_none, args, kwargs):
        wrapped = args[0]
        if not inspect.isclass(superclass):
            raise TypeError("@ensure_decorates_class() arg 1 must be a class")

        if not inspect.isclass(wrapped):
            raise RuntimeError(f"Decorator @{decorator.__name__} must be "
                               "applied to a class, not to a "
                               f"<{wrapped.__class__.__name__}>")

        if not issubclass(wrapped, superclass):
            raise RuntimeError(f"Decorator @{decorator.__name__} must be "
                               f"applied to a {superclass.__name__} subclass, "
                               f"not to a {wrapped.__name__} class.")

        return decorator(*args, **kwargs)
    return _wrapper


# The following decorators are useful for development, but should be removed
# otherwise when normally running


def profile_calls(sort_args=['cumulative'], print_args=[10]):
    """
    Runs a profiler on the wrapped function, using cProfile. This profiles
    exclusively function calls.
    """
    profiler = cProfile.Profile()

    def decorator(func):
        def inner(*args, **kwargs):
            result = None
            try:
                profiler.enable()
                result = func(*args, **kwargs)
                profiler.disable()
            finally:
                stats = pstats.Stats(profiler)
                print("################",
                      f"Profiling funtion {func.__name__}",
                      "################")
                stats.sort_stats(*sort_args).print_stats(*print_args)
            return result
        return inner
    return decorator


def profile_lines(func):
    """Profile execution time of each line of the decorated function."""
    try:
        LineProfiler
    except NameError:
        raise RuntimeError("No line_profiler module found.") from None

    def _wrapper(*args, **kwargs):
        profiler = LineProfiler()
        profiled_func = profiler(func)
        try:
            result = profiled_func(*args, **kwargs)
        finally:
            profiler.print_stats()
        return result
    return _wrapper


def exec_time(func):
    """
    Measures execution time of the wrapped function
    """
    def _wrapper(*args, **kwargs):
        t0 = timer()
        result = func(*args, **kwargs)
        dt = timer()-t0
        if dt > 1:
            mult = 1
            unit = ''
        elif dt > 1e-3:
            mult = 1e3
            unit = 'm'
        elif dt > 1e-6:
            mult = 1e6
            unit = 'u'
        else:
            mult = 1e9
            unit = 'n'
        fname = str(func).replace('<','').split(' at 0x')[0]
        print(f"Execution time of {fname}: {dt*mult:.1f} {unit}s")
        return result
    return _wrapper




"""Section Reference Calculation."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-11'
__license__ = 'GPLv3+'

import copy
import logging
import os
from pathlib import Path
import shutil
import subprocess

import fortranformat as ff
import numpy as np

from viperleed.calc.files import beams
from viperleed.calc.files import iorefcalc
from viperleed.calc.files import parameters
from viperleed.calc.files.ivplot import plot_iv
from viperleed.calc.lib import leedbase
from viperleed.calc.lib import parallelization
from viperleed.calc.lib.base import splitMaxRight
from viperleed.calc.lib.checksums import validate_multiple_files
from viperleed.calc.lib.version import Version

logger = logging.getLogger(__name__)

# TODO: we should have a parent class for compile tasks (issue #43)
# CompileTask subclasses would need a class-level list of
# glob patterns for retrieving source files. Probably allowing
# a special "{__sourcedir__}" format specifier to be formatted
# with .format(__sourcedir__=self.source_dir) before globbing.
# Similar considerations regarding the base names for foldername
# and exename.
# TODO: when implementing the compile task class, we should also
# refactor the tenserleed source class such that we can directly use
# zipped source directories
class RefcalcCompileTask():
    """Stores information for a worker to compile a refcalc file, and keeps
    track of the folder that the compiled file is in afterwards."""

    def __init__(self, param, lmax, fortran_comp, sourcedir,
                 basedir=Path()):
        self.param = param
        self.lmax = lmax
        self.fortran_comp = fortran_comp
        self.source_dir = Path(sourcedir).resolve()  # where the fortran files are
        self.basedir = Path(basedir)  # where the calculation is based
        self.foldername = f'refcalc-compile_LMAX{lmax}'
        self.exename = f'refcalc-{lmax}'

        if os.name == 'nt':
            self.exename += '.exe'

    @property
    def logfile(self):
        return self.basedir / self.foldername / "fortran-compile.log"

    @property
    def compile_log_name(self):
        # name as it should appear in the compile_logs directory
        return self.foldername

    def get_source_files(self):
        """Return a tuple of source files needed for running a refcalc."""
        sourcedir = Path(self.source_dir).resolve()
        libpath = sourcedir / 'lib'
        srcpath = sourcedir / 'src'
        lib_tleed = next(libpath.glob('lib.tleed*'), None)
        srcname = next(srcpath.glob('ref-calc*'), None)
        globalname = srcpath / "GLOBAL"
        _muftin = self.basedir / "muftin.f"
        muftinname =_muftin if _muftin.is_file() else None
        if any(f is None for f in (srcname, lib_tleed)):
            raise RuntimeError(f"Source files missing in {sourcedir}")          # TODO: use a more appropriate custom exception in CompileTask (e.g., MissingSourceFileError)
        return lib_tleed, srcname, globalname, muftinname

    def copy_source_files_to_local(self):
        """Copy ref-calc files to current directory."""
        for filepath in self.get_source_files():
            if filepath:
                shutil.copy2(filepath, filepath.name)


class RefcalcRunTask():
    """Stores information for a worker to create a subfolder, copy input there,
    compile and run a reference calculation, and copy results back."""

    def __init__(self, fin, energy, comptask, logname,
                 collect_at="", single_threaded=False, tl_version=0.):
        self.fin = fin
        self.energy = energy
        self.comptask = comptask
        self.logname = logname
        self.foldername = "refcalc-part_{:.2f}eV".format(energy)
        self.collect_at = collect_at
        self.single_threaded = single_threaded
        self.tl_version = tl_version


def compile_refcalc(comptask):
    """Function meant to be executed by parallelized workers. Executes a
    RefcalcCompileTask."""
    workfolder = os.path.join(comptask.basedir, comptask.foldername)
    # make folder and go there:
    if os.path.isdir(workfolder):
        logger.warning("Folder "+comptask.foldername+" already exists. "
                       "Contents may get overwritten.")
    else:
        os.mkdir(workfolder)
    os.chdir(workfolder)
    # write PARAM:
    try:
        with open("PARAM", "w") as wf:
            wf.write(comptask.param)
    except Exception:
        logger.error("Error writing PARAM file: ", exc_info=True)
        return ("Error encountered by RefcalcCompileTask "
                + comptask.foldername + "while trying to write PARAM file.")

    try:
        comptask.copy_source_files_to_local()
    except Exception:
        logger.error("Error getting TensErLEED files for "
                     "refcalc-amplitudes: ", exc_info=True)
        return ("Error encountered by RefcalcCompileTask "
                + comptask.foldername + " while trying to fetch fortran "
                "source files")

    # TODO: we could skip this, if we implemented a general CompileTask (Issue #43)
    (libname, srcname,
     _, muftinname) = (
         str(fname.name) if fname is not None else None
         for fname in comptask.get_source_files()
         )

    compile_list = [(libname, "lib.tleed.o"), (srcname, "main.o")]
    if muftinname:
        compile_list.append((muftinname, "muftin.o"))
    # compile
    ctasks = [(comptask.fortran_comp[0] + " -o " + oname + " -c",
               fname, comptask.fortran_comp[1]) for (fname, oname)
              in compile_list]
    ctasks.append((comptask.fortran_comp[0] + " -o " + comptask.exename,
                   " ".join(list(zip(*compile_list))[1]),
                   comptask.fortran_comp[1]))
    try:
        leedbase.fortran_compile_batch(ctasks)
    except Exception as e:
        logger.error("Error compiling fortran files: " + str(e))
        return ("Fortran compile error in RefcalcCompileTask "
                + comptask.foldername)
    os.chdir(comptask.basedir)
    return ""


def run_refcalc(runtask):
    """Runs a part of a reference calculation in a subfolder, or the whole
    refcalc here if in single-threaded mode."""
    base = runtask.comptask.basedir
    workfolder = base
    task_name = "(single-threaded)"
    if not runtask.single_threaded:
        # make folder and go there:
        workfolder = os.path.join(base, runtask.foldername)
        task_name = runtask.foldername
        if os.path.isdir(workfolder):
            logger.warning("Folder "+runtask.foldername+" already exists. "
                           "Contents may get overwritten.")
        else:
            os.mkdir(workfolder)
        os.chdir(workfolder)

    if runtask.single_threaded:
        fin = runtask.fin
        logname = runtask.logname
    else:
        logname = "refcalc.log"
        fin = edit_fin_energy_lmax(runtask)
        try:
            with open("refcalc-FIN", "w") as wf:
                wf.write(fin)
        except Exception:
            pass  # local FIN is just for information...

    # get executable
    exename = runtask.comptask.exename
    try:
        shutil.copy2(os.path.join(base, runtask.comptask.foldername, exename),
                     os.path.join(workfolder, exename))
    except Exception:
        logger.error("Error getting refcalc executable: ", exc_info=True)
        return ("Error encountered by RefcalcRunTask " + task_name
                + ": Failed to get refcalc executable.")
    # run execution
    try:
        with open(logname, "w") as log:
            subprocess.run(os.path.join(workfolder, exename),
                           input=fin, encoding="ascii",
                           stdout=log, stderr=log)
    except Exception:
        logger.error("Error while executing reference calculation "
                     + task_name + ". Also check refcalc log file.",
                     exc_info=True)
        return ("Error encountered by RefcalcRunTask " + task_name
                + ": Error during refcalc execution.")

    if runtask.single_threaded:
        return ""

    # move/copy files out
    if runtask.collect_at:
        targetpath = os.path.abspath(runtask.collect_at)
    else:
        targetpath = base
    en_str = "_{:.2f}eV".format(runtask.energy)
    tensorfiles = [f for f in os.listdir() if f.startswith("T_")
                   and os.path.isfile(f)]
    for tf in tensorfiles:
        try:   # move instead of copy to not duplicate the large files
            shutil.move(os.path.join(workfolder, tf),
                        os.path.join(targetpath, tf + en_str))
        except Exception:
            logger.error("Failed to copy refcalc output file " + tf +
                         " to main folder.", exc_info=True)
            return ("Error encountered by RefcalcRunTask " + task_name
                    + ": Failed to copy Tensor file out.")
    try:
        shutil.copy2(os.path.join(workfolder, "fd.out"),
                     os.path.join(targetpath, "fd" + en_str + ".out"))
    except Exception:
        logger.error("Failed to copy refcalc output file fd.out "
                     " to main folder.", exc_info=True)
        return ("Error encountered by RefcalcRunTask " + task_name
                + ": Failed to copy fd.out file out.")
    try:
        shutil.copy2(os.path.join(workfolder, "amp.out"),
                     os.path.join(targetpath, "amp" + en_str + ".out"))
    except FileNotFoundError:
        if runtask.tl_version >= Version('1.7.3'):
            logger.warning("Refcalc output file amp.out not found.")
    except Exception as e:      # warn but continue
        logger.warning("Failed to copy refcalc output file amp.out "
                       "to main folder: " + str(e))
    # append log
    log = ""
    try:
        with open(logname, "r") as rf:
            log = rf.read()
    except Exception:
        logger.warning("Could not read local refcalc log " + task_name)
    if log != "":
        globallog = os.path.join(base, runtask.logname)
        try:
            with open(globallog, "a") as wf:
                wf.write("\n\n### STARTING LOG FOR " + task_name
                         + " ###\n\n" + log)
        except Exception:
            logger.warning("Error writing refcalc log part "
                           + task_name + ": ", exc_info=True)
    # clean up
    os.chdir(base)
    try:
        shutil.rmtree(workfolder)
    except Exception:
        logger.warning("Error deleting folder " + runtask.foldername)
    return ""


def edit_fin_energy_lmax(runtask):
    """modify FIN: replace the energy range (second line)"""
    comment, _, rest = runtask.fin.split("\n", maxsplit=2)
    if runtask.tl_version < Version('1.7.0'):
        eformatter = ff.FortranRecordWriter('3F7.2')
        lj = 24
    else:
        eformatter = ff.FortranRecordWriter('3F9.2')
        lj = 30
    energy = (eformatter.write([runtask.energy, runtask.energy + 0.01, 1.0])
              .ljust(lj) + 'EI,EF,DE')
    # now replace LMAX - now makes sure LMAX can be part of directory name
    # this works because even if the directory were to be named LMAX, there is
    #   a timestap after it rather than a \n
    before_LMAX, after_LMAX = rest.split("   LMAX", maxsplit=1)
    before_LMAX = splitMaxRight(before_LMAX, "\n")[0]
    after_LMAX = (str(runtask.comptask.lmax).rjust(3).ljust(45)
                  + "LMAX" + after_LMAX)
    # fin = finparts[0] + "\n" + nl + finparts[1]
    fin = "\n".join((comment, energy, before_LMAX, after_LMAX))
    return fin


def refcalc(sl, rp, subdomain=False, parent_dir=Path()):
    """Main function to execute the reference calculation segment."""
    if rp.domainParams:
        refcalc_domains(rp)
        return
    sl.update_cartesian_from_fractional(update_origin=True)
    sl.update_layer_coordinates()
    # delete old refcalc-fd.out if present - not earlier because can be input
    #   for r-factor calculation if no refcalc is executed
    try:
        os.remove("refcalc-fd.out")
    except FileNotFoundError:
        pass
    try:
        iorefcalc.writeAUXLATGEO(sl, rp)
    except Exception:
        logger.error("Exception during writeAUXLATGEO: ")
        raise
    try:
        iorefcalc.writeAUXNONSTRUCT(sl, rp)
    except Exception:
        logger.error("Exception during writeAUXNONSTRUCT: ")
        raise
    if rp.TL_VERSION < Version('1.7.3'):
        try:
            beams.writeAUXBEAMS(ivbeams=rp.ivbeams, beamlist=rp.beamlist)
        except Exception:
            logger.error("Exception during writeAUXBEAMS: ")
            raise
    try:
        iorefcalc.writeAUXGEO(sl, rp)
    except Exception:
        logger.error("Exception during writeAUXGEO: ")
        raise
    try:
        fin = iorefcalc.collectFIN(version=rp.TL_VERSION)
    except Exception:
        logger.error("Exception while trying to collect input for "
                     "refcalc FIN: ")
        raise
    try:
        with open("refcalc-FIN", "w") as wf:
            wf.write(fin)
        logger.debug("Wrote input for refcalc as file refcalc-FIN.")
    except Exception:
        logger.error(
            "Exception while trying to write refcalc-FIN file. "
            "Execution will proceed. The exception was: ",
            exc_info=True
            )
    if rp.TL_VERSION < Version('1.7.0'):   # muftin.f deprecated in version 1.7
        try:
            iorefcalc.writeMuftin(rp)
        except Exception:
            logger.error("Exception during writeMuftin: ")
            raise
    if rp.SUPPRESS_EXECUTION:
        logger.warning("SUPPRESS_EXECUTION parameter is on. Reference "
                       "calculation will not proceed. Stopping...")
        rp.setHaltingLevel(3)
        return

    energies = np.arange(rp.THEO_ENERGIES.start, rp.THEO_ENERGIES.stop+0.01,         # TODO: use better arange
                         rp.THEO_ENERGIES.step)
    tl_source = rp.get_tenserleed_directory()
    tl_path = tl_source.path
    rp.updateCores()
    single_threaded = (rp.N_CORES <= 1)
    if rp.FORTRAN_COMP[0] == "":
        try:
            rp.getFortranComp()
        except Exception:
            logger.error("No fortran compiler found, cancelling...")
            raise RuntimeError("Fortran compile error")

    # first, figure out for which LMAX to compile:
    if single_threaded or rp.LMAX.has_single_value or rp.TL_VERSION <= Version('1.6'):
        which_lmax = {rp.LMAX.max,}
    else:    # find appropriate LMAX per energy
        ps_en = [(i, ps[0]*leedbase.HARTREE_TO_EV) for (i, ps) in enumerate(rp.phaseshifts)]
        lmax = {}  # lmax as a function of energy
        warn_small = True
        warn_large = True
        _, _max = rp.get_limits("LMAX")
        for en in energies:
            try:
                ps_ind = [pe[0] for pe in ps_en if pe[1] >= en][0]
            except IndexError:
                if ps_en[-1][1] > en and ps_en[-1][1] - en > 1.:
                    logger.warning(
                        "No approriate phaseshifts found for {:.2f} eV. Will "
                        "try highest available phaseshift energy {:.2f} eV "
                        "instead.".format(en, ps_en[-1][1]))
                ps_ind = len(rp.phaseshifts) - 1
            lmax_cands = [1]
            for site_ps in rp.phaseshifts[ps_ind][1]:
                try:
                    lmax_cands.append(max(
                        [site_ps.index(v) for v in site_ps
                         if abs(v) > rp.PHASESHIFT_EPS]) + 1)
                except (IndexError, ValueError):
                    pass
            lmax[en] = min(max((rp.LMAX.min, *lmax_cands)), rp.LMAX.max)
            if lmax[en] < 6 and warn_small:
                warn_small = False
                logger.debug("Found small LMAX value based on PHASESHIFT_EPS "
                             f"parameter (LMAX = {lmax[en]}, E = {en:.2f} eV)")
            if lmax[en] > _max:
                lmax[en] = _max
                if warn_large:
                    warn_large = False
                    logger.info("The LMAX found based on the PHASESHIFT_EPS "
                                f"parameter is greater than {_max}, which is "
                                "currently not supported. LMAX was set to "
                                f"{_max}.")
        which_lmax = set(lmax.values())

    # collect compile tasks
    comp_tasks = []
    collect_param = ""
    for lm in which_lmax:
        try:
            param = iorefcalc.writePARAM(sl, rp, lmax=lm)
        except Exception:
            logger.error("Exception during writePARAM: ",
                         exc_info=rp.is_debug_mode)
            raise
        comp_tasks.append(RefcalcCompileTask(param, lm, rp.FORTRAN_COMP,
                                             tl_path, basedir=rp.workdir))
        collect_param += f"### PARAM file for LMAX = {lm} ###\n\n{param}\n\n"
    try:
        with open("refcalc-PARAM", "w") as wf:
            wf.write(collect_param)
    except Exception:
        # just for information, can continue
        logger.warning("Error writing refcalc-PARAM file: ", exc_info=True)

    # set up log
    logname = "refcalc-"+rp.timestamp+".log"
    if not single_threaded:
        try:
            with open(logname, "w") as wf:
                wf.write(
                    "Logs from multiple reference calculations are "
                    "collected  here. Their order may not be preserved.\n")
        except Exception:
            logger.warning("Error creating refcalc log file. This will not "
                           "affect execution, proceeding...")
    # set up collection directory
    if not single_threaded:
        collection_dir = os.path.join(os.getcwd(), "refcalc-out")
        if os.path.isdir(collection_dir):
            try:
                shutil.rmtree(collection_dir)
            except Exception:
                logger.warning(
                    "Failed to delete existing folder "
                    + os.path.basename(collection_dir) + ". This may cause "
                    "old data to end up in the final Tensors, check results!")
        os.makedirs(collection_dir, exist_ok=True)
    # collect run tasks
    ref_tasks = []
    if not single_threaded:
        for en in sorted(energies, reverse=True):
            if len(which_lmax) == 1:
                ct = comp_tasks[0]
            else:
                ct = next(ct for ct in comp_tasks if ct.lmax == lmax[en])
            ref_tasks.append(RefcalcRunTask(fin, en, ct, logname,
                                            collect_at=collection_dir,
                                            single_threaded=False,
                                            tl_version=rp.TL_VERSION))
    else:
        ct = comp_tasks[0]
        ref_tasks.append(RefcalcRunTask(fin, -1, ct, logname,
                                        single_threaded=True,
                                        tl_version=rp.TL_VERSION))

    # Validate TensErLEED checksums
    if not rp.TL_IGNORE_CHECKSUM:
        # @issue #43: this could be a class method
        files_to_check = (file for file in comp_tasks[0].get_source_files()
                          if file and 'muftin' not in file.name.lower())
        validate_multiple_files(files_to_check,
                                logger, "reference calculation",
                                rp.TL_VERSION)

    if single_threaded:
        home = os.getcwd()
        try:
            r = compile_refcalc(comp_tasks[0])
        except Exception:
            # if something goes wrong copy log file to compile logs
            leedbase.copy_compile_log(rp, comp_tasks[0].logfile,
                                      comp_tasks[0].compile_log_name)
            raise
        finally:
            os.chdir(home)
        if r:
            logger.error(r)
            raise RuntimeError("Error compiling fortran files.")
        logger.info("Compiling fortran files...")
        logger.info("Starting reference calculation...\n"
                    "Refcalc log will be written to file "+logname)
        logger.info("Reference calculation running without parallelization. "
                    "Set the N_CORES parameter to speed it up.")
        try:
            r = run_refcalc(ref_tasks[0])
        finally:
            os.chdir(home)
        if r:
            logger.error(r)
            raise RuntimeError("Error in reference calculation.")
        logger.info("Reference calculation finished. Processing files...")
    else:
        # compile files
        logger.info("Compiling fortran files...")
        poolsize = min(len(comp_tasks), rp.N_CORES)
        try:
            parallelization.monitoredPool(rp, poolsize,
                                          compile_refcalc,
                                          comp_tasks,
                                          update_from=parent_dir)
        except Exception:
            # save log files in case of error:
            for ct in comp_tasks:
                leedbase.copy_compile_log(rp, ct.logfile, ct.compile_log_name)
            raise
        if rp.STOP:
            return
        # run executions
        logger.info("Running reference calculations...")
        poolsize = min(len(ref_tasks), rp.N_CORES)
        parallelization.monitoredPool(rp, poolsize, run_refcalc, ref_tasks,
                                      update_from=parent_dir)
        if rp.STOP:
            return
        logger.info("Reference calculations finished. Processing files...")

    # clean up compile folders - AMI: move logs first to compile_logs !
    for ct in comp_tasks:
        leedbase.copy_compile_log(rp, ct.logfile, ct.compile_log_name)
        try:
            shutil.rmtree(os.path.join(ct.basedir, ct.foldername))
        except Exception:
            logger.warning("Error deleting refcalc compile folder "
                           + ct.foldername)

    if not single_threaded:
        iorefcalc.combine_fdout(oripath=collection_dir)
        if 1 in rp.TENSOR_OUTPUT:
            iorefcalc.combine_tensors(oripath=collection_dir)
        try:
            shutil.rmtree(collection_dir)
        except Exception:
            logger.warning("Failed to delete empty directory "
                           + os.path.basename(collection_dir))

    try:
        rp.theobeams["refcalc"], rp.refcalc_fdout = iorefcalc.readFdOut()
    except FileNotFoundError:
        logger.error("fd.out not found after reference calculation. "
                     "Check settings and refcalc log.")
        raise
    except Exception:
        logger.error("Error reading fd.out after reference calculation. "
                     "Check settings and refcalc log.")
        raise
    if rp.theobeams["refcalc"] is None:
        logger.error("No data found in fd.out. Check if file is empty.")
        raise RuntimeError                                                      # TODO: better exception
    # clear oriState for atoms and sites, current state will be new origin
    for at in sl:
        at.oriState = None
    for site in sl.sitelist:
        site.oriState = None
    # compare beam sets:
    eq = True
    eps = 1e-3
    if len(rp.ivbeams) != len(rp.theobeams["refcalc"]):
        eq = False
        message = "Number of beams is inconsitent."
    else:
        eq = all([rp.ivbeams[i].isEqual(rp.theobeams["refcalc"][i], eps=eps)
                  for i in range(0, len(rp.ivbeams))])
        message = "Beam labels are inconsistent."
    if not eq:
        logger.error("The list of beams read from IVBEAMS is not "
                     "equivalent to the list of beams in the fd.out file "
                     "produced by the reference calculation! " + message)
        rp.setHaltingLevel(2)
    # check for beams with very low values
    if not subdomain:
        small_intensities = []
        for b in [b for b in rp.theobeams["refcalc"]
                  if max(b.intens.values()) < 1e-10]:
            small_intensities.append(b.label)
        if small_intensities:
            logger.warning(
                "Some calculated beams only contain very small intensities. "
                "This may indicate that the beams do not exist for this "
                "structure. Consider removing them from IVBEAMS: "
                + ", ".join(small_intensities))
    try:
        beams.writeOUTBEAMS(rp.theobeams["refcalc"], filename="THEOBEAMS.csv")
        theobeams_norm = copy.deepcopy(rp.theobeams["refcalc"])
        for b in theobeams_norm:
            b.normMax()
        beams.writeOUTBEAMS(theobeams_norm, filename="THEOBEAMS_norm.csv")
    except Exception:
        logger.error("Error writing THEOBEAMS after reference "
                     "calculation: ", exc_info=True)
        rp.setHaltingLevel(2)
    if len(rp.theobeams["refcalc"][0].complex_amplitude) != 0:
        try:
            beams.writeOUTBEAMS(rp.theobeams["refcalc"],
                                filename="Complex_amplitudes_real.csv",
                                which="amp_real")
            beams.writeOUTBEAMS(rp.theobeams["refcalc"],
                                filename="Complex_amplitudes_imag.csv",
                                which="amp_imag")
        except Exception:
            logger.error("Error writing complex amplitudes after reference "
                         "calculation.", exc_info=rp.is_debug_mode)
    try:
        plot_iv(theobeams_norm, "THEOBEAMS.pdf", formatting=rp.PLOT_IV)
    except Exception:
        logger.warning("Error writing THEOBEAMS.pdf after reference "
                       "calculation.")

    try:
        os.rename('fd.out', 'refcalc-fd.out')
    except Exception:
        logger.warning("Failed to rename refcalc output file fd.out to "
                       "refcalc-fd.out")
    try:
        os.rename('amp.out', 'refcalc-amp.out')
    except FileNotFoundError:
        pass
    except Exception:
        logger.warning("Failed to rename refcalc output file amp.out to "
                       "refcalc-amp.out")
    if 1 not in rp.TENSOR_OUTPUT:
        return
    # move and zip tensor files
    rp.TENSOR_INDEX = leedbase.getMaxTensorIndex() + 1
    if "Tensors" not in rp.manifest:
        rp.manifest.append("Tensors")
    dn = "Tensors_"+str(rp.TENSOR_INDEX).zfill(3)
    os.makedirs(os.path.join(".", "Tensors", dn), exist_ok=True)
    try:
        for tf in [f for f in os.listdir('.') if f.startswith("T_")]:
            shutil.move(tf, os.path.join(".", "Tensors", dn, tf))
    except Exception:
        logger.error("Error moving Tensor files: ")
        raise
    tInputFiles = ["POSCAR", "PARAMETERS", "VIBROCC", "IVBEAMS",
                   "PHASESHIFTS"]
    for f in [f for f in tInputFiles if f in os.listdir('.')]:
        of = f
        for fn in ["POSCAR", "VIBROCC"]:
            if (f == fn and 3 in rp.runHistory
                    and os.path.isfile(fn + "_OUT_" + rp.timestamp)):
                of = fn + "_OUT_" + rp.timestamp
        try:
            shutil.copy2(of, os.path.join("Tensors", dn, f))
        except Exception:
            logger.warning("Failed to add input file " + f
                           + " to Tensors folder.")
    try:
        shutil.copy2("refcalc-fd.out",
                     os.path.join("Tensors", dn, "refcalc-fd.out"))
    except Exception:
        logger.warning("Failed to copy refcalc-fd.out to Tensors folder.")
    # modify PARAMETERS to contain the energies and LMAX that were really used
    if os.path.isfile(os.path.join("Tensors", dn, "PARAMETERS")):
        parameters.modify(rp, "THEO_ENERGIES",
                          path=os.path.join("Tensors", dn), suppress_ori=True)
        parameters.modify(rp, "LMAX",
                          path=os.path.join("Tensors", dn), suppress_ori=True)

    # remove references to Deltas from old tensors
    _reinitialize_deltas(rp, sl)

    return


def _reinitialize_deltas(param, slab):
    """Removes references to deltas from previous tensors.
    Delete old delta files in main work folder, if necessary.
    (there should not be any, unless there was an error)
    Also empty all atom.known_deltas because they would refer to
    previous tensors.

    Parameters
    ----------
    param : Rparam
    slab : Slab
    """
    # delete old delta files in main work folder, if necessary
    for df in [f for f in os.listdir(param.workdir) if f.startswith("DEL_") and
               os.path.isfile(param.workdir / f)]:
        try:
            os.remove(df)
        except Exception:
            logger.warning(
                "Error deleting old Delta file in work directory. This may "
                "cause the delta file to incorrectly be labelled as belonging "
                "with the new set of tensors.")

    # empty atom.known_deltas
    for at in slab:
        at.known_deltas = []


def runDomainRefcalc(dp):
    """Runs the reference calculation for one domain, based on the
    DomainParameters object."""
    home = os.getcwd()
    try:
        os.chdir(dp.workdir)
        refcalc(dp.sl, dp.rp, subdomain=True)
    except Exception:
        logger.error("Exception during reference calculation for domain {}: "
                     .format(dp.name), exc_info=True)
        raise
    finally:
        os.chdir(home)
    return


def refcalc_domains(rp):
    """Runs reference calculations for the domains that require them."""
    rr = [dp for dp in rp.domainParams if dp.refcalcRequired]
    if not rr:
        logger.info("Found no domain which requires a reference calculation.")
        return
    # make sure there's a compiler ready, and we know the number of cores:
    if rp.FORTRAN_COMP[0] == "":
        try:
            rp.getFortranComp()
        except Exception:
            logger.error("No fortran compiler found, cancelling...")
            raise
    for dp in rp.domainParams:
        dp.rp.FORTRAN_COMP = rp.FORTRAN_COMP
    rp.updateCores()  # if number of cores is not defined, try to find it
    rr = [dp for dp in rp.domainParams if dp.refcalcRequired]
    logger.info("Running reference calculations in subfolders for domains: "
                + ", ".join([d.name for d in rr]))

    for dp in rr:
        logger.info("Starting reference calculation for domain {}"
                    .format(dp.name))
        runDomainRefcalc(dp)
    logger.info("Domain reference calculations finished.")

    if len(rr) < len(rp.domainParams):
        return
    # if refcalcs were done for all domains, get averaged beams
    if 3 in rp.runHistory and rp.searchResultConfig:
        weights = [rp.searchResultConfig[0][i][0]
                   for i in range(0, len(rp.domainParams))]
        logger.info(
            "Reference calculations were done for all domains. Getting "
            "weighted average over domain beams, using weights from last "
            "search result...")
    else:
        weights = None
        logger.info(
            "Reference calculations were done for all domains, but no "
            "area weights for the different domains are available yet. "
            "Getting unweighted average over domain beams...")
    rp.theobeams["refcalc"] = beams.averageBeams([
        dp.rp.theobeams["refcalc"] for dp in rp.domainParams], weights=weights)
    try:
        beams.writeOUTBEAMS(rp.theobeams["refcalc"], filename="THEOBEAMS.csv")
        theobeams_norm = copy.deepcopy(rp.theobeams["refcalc"])
        for b in theobeams_norm:
            b.normMax()
        beams.writeOUTBEAMS(theobeams_norm, filename="THEOBEAMS_norm.csv")
    except Exception:
        logger.error("Error writing THEOBEAMS after reference calculation.",
                     exc_info=rp.is_debug_mode)
    try:
        rp.superpos_specout = beams.writeFdOut(rp.theobeams["refcalc"],
                                               rp.beamlist,
                                               filename="refcalc-fd.out",
                                               header=rp.systemName)
    except Exception:
        logger.error("Error writing averaged refcalc-fd.out for R-factor "
                     "calculation.", exc_info=rp.is_debug_mode)



"""Module gui.py of viperleed.

======================================
  ViPErLEED Graphical User Interface
======================================

Author: Michele Riva
Created: 2020-01-11

This is the module that invokes the execution of the
Graphical User Interface
"""

import sys
import os

from viperleed import GLOBALS

try:
    import PyQt5.QtCore as qtc
except ImportError:
    GLOBALS['USE_GUI'] = False
else:
    import PyQt5.QtGui as qtg
    import PyQt5.QtWidgets as qtw
    GLOBALS['USE_GUI'] = True

from viperleed import guilib as gl
from viperleed.cli_base import ViPErLEEDCLI


class ViPErLEEDGUICLI(ViPErLEEDCLI, cli_name='gui'):
    """Main entry point for GUI. Both command-line and graphical."""

    def __call__(self, args=None):
        """Call either the CLI or graphical versions of the GUI."""
        if is_commandline_mode():
            return commandline_main()
            _ = super().__call__(args)                                          # TODO: This line is unreachable because we don't handle the gui arguments correctly yet
        return gui_main()


def is_commandline_mode():
    """Return whether the system requires to run in command line mode.

    This is the case if, e.g., the correct modules are not installed)
    or if the user asked to run in command line mode via the --nogui
    command line argument.
    """
    needs_commandline = not GLOBALS['USE_GUI']
    needs_commandline |= gl.BACKEND is None
    needs_commandline |= '--nogui' in sys.argv
    return needs_commandline


def commandline_main():
    """Body of the command-line version of the GUI.

    In this case, there is very little graphics involved.
    """
    print('Running in command line version...', flush=True, end='')
    print('Not implemented yet.', flush=True)


def resources_path(dir_name):
    """Return the correct path to dir_name.

    This is useful when building an executable from pyinstaller.
    When built from pyinstaller, it takes the path relative to the
    temporary path in which the "exe" is extracted during execution.

    Parameters
    ----------
    dir_name : str
        Path relative to the one from which resources_path is called.

    Returns
    -------
    str
        Modified path, only if the GUI is running from a pyinstaller
        'executable', otherwise returns dir_name unchanged.
    """
    # EVENTUALLY IT IS PROBABLY BETTER TO INCLUDE THE WHOLE /fonts FOLDER from
    # '/guilib' IN THE CORRECT PLACE, AND HAVE resources_path RETURN ITS BASE
    # PATH (i.e., the top-level folder in which the exe is) IF PYINSTALLER IS
    # USED
    if hasattr(sys, '_MEIPASS'):
        return os.path.join(sys._MEIPASS, dir_name)
    return dir_name


def gui_main():
    """Body of the GUI when running with graphics capability.

    Body of the functionality that invokes the ViPErLEED
    Graphical User Interface.
    """
    gl.catch_gui_crash()

    print('Loading GUI...', flush=True, end='')
    qtg.QGuiApplication.setAttribute(qtc.Qt.AA_UseHighDpiPixmaps)
    app = qtw.QApplication(sys.argv)

    # Import some fonts from ./fonts folder
    font_path = resources_path('guilib/fonts')
    # * Text: family =  'DejaVu Sans'
    qtg.QFontDatabase.addApplicationFont(os.path.join(font_path,
                                                      'DejaVuSans.ttf'))
    # * Math: family =  'CMU Serif'
    qtg.QFontDatabase.addApplicationFont(os.path.join(font_path, 'cmunrm.otf'))

    # leed_gui = gl.LEED_GUI()
    # leed_gui.show()
    gl.show_use_betatest_version_popup()

    print('Done', flush=True)

    # sys.exit(app.exec_())
    sys.exit()


if __name__ == '__main__':
    ViPErLEEDGUICLI.run_as_script()



"""
======================================
  ViPErLEED Graphical User Interface
======================================
 *** module guilib.leedsim.classes ***

Contains Qt-independent classes that are used by widgets to display the real
space lattices and the LEED pattern. Used to be a single module classes.py

Author: Michele Riva
Created: 2021-03-13
"""

from viperleed.guilib.leedsim.classes.woods import Woods
from viperleed.guilib.leedsim.classes.realspace import RealSpace
from viperleed.guilib.leedsim.classes.beams import LEEDEquivalentBeams
from viperleed.guilib.leedsim.classes.symdomains import LEEDSymmetryDomains
from viperleed.guilib.leedsim.classes.leedpattern import (LEEDStructuralDomains,
                                                          # LEEDPattern,
                                                          # LEEDsubpattern
                                                          )
from viperleed.guilib.leedsim.classes.oldleedpatterns import (LEEDPattern,  # These will be replaced once the new LEEDPattern/LEEDsubpattern work as they should
                                                              LEEDsubpattern)


